"""
Keeps information about ANAC competitions and Genius Agents
"""
from __future__ import annotations

import itertools

__all__ = [
    "GENIUS_INFO",
    "AGENT_BASED_NEGOTIATORS",
    "PARTY_BASED_NEGOTIATORS",
    "ALL_NEGOTIATORS",
    "TEST_FAILING_NEGOTIATORS",
    "ALL_PASSING_NEGOTIATORS",
    "TESTED_NEGOTIATORS",
    "ALL_GENIUS_BASIC_NEGOTIATORS",
    "ALL_GENIUS_INVALID_NEGOTIATORS",
    "ALL_GENIUS_NEGOTIATORS",
    "get_anac_agents",
]
GENIUS_INFO = {
    2010: {
        "winners": [
            [("AgentK", "agents.anac.y2010.AgentK.Agent_K")],
            [("Yushu", "agents.anac.y2010.Yushu.Yushu")],
            [("Nozomi", "agents.anac.y2010.Nozomi.Nozomi")],
            [("IAMhaggler", "agents.anac.y2010.Southampton.IAMhaggler")],
        ],
        "finalists": [
            [("AgentK", "agents.anac.y2010.AgentK.Agent_K")],
            [("Yushu", "agents.anac.y2010.Yushu.Yushu")],
            [("Nozomi", "agents.anac.y2010.Nozomi.Nozomi")],
            [("IAMhaggler", "agents.anac.y2010.Southampton.IAMhaggler")],
        ],
        "linear": True,
        "learning": False,
        "multilateral": False,
        "bilateral": True,
        "reservation": False,
        "discounting": True,
        "uncertainty": False,
        "elicitation": False,
    },
    2011: {
        "winners": [
            [("HardHeaded", "agents.anac.y2011.HardHeaded.KLH")],
            [("Gahboninho", "agents.anac.y2011.Gahboninho.Gahboninho")],
            [("IAMhaggler2011", "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011")],
        ],
        "finalists": [
            [("HardHeaded", "agents.anac.y2011.HardHeaded.KLH")],
            [("Gahboninho", "agents.anac.y2011.Gahboninho.Gahboninho")],
            [("IAMhaggler2011", "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011")],
            [("BRAMAgent", "agents.anac.y2011.BramAgent.BRAMAgent")],
            [("AgentK2", "agents.anac.y2011.AgentK2.Agent_K2")],
            [("TheNegotiator", "agents.anac.y2011.TheNegotiator.TheNegotiator")],
            [("NiceTitForTat", "agents.anac.y2011.Nice_Tit_for_Tat.NiceTitForTat")],
            [("ValueModelAgent", "agents.anac.y2011.ValueModelAgent.ValueModelAgent")],
        ],
        "linear": True,
        "learning": False,
        "multilateral": False,
        "bilateral": True,
        "reservation": False,
        "discounting": True,
        "uncertainty": False,
        "elicitation": False,
    },
    2012: {
        "winners": [
            [
                (
                    "CUHKAgent",
                    "agents.anac.y2012.CUHKAgent.CUHKAgent",
                )
            ],
            [("AgentLG", "agents.anac.y2012.AgentLG.AgentLG")],
            [
                (
                    "OMACagent",
                    "agents.anac.y2012.OMACagent.OMACagent",
                ),
                (
                    "TheNegotiatorReloaded",
                    "agents.anac.y2012.TheNegotiatorReloaded.TheNegotiatorReloaded",
                ),
            ],
        ],
        "best_discounted": [
            (
                "CUHKAgent",
                "agents.anac.y2012.CUHKAgent.CUHKAgent",
            )
        ],
        "best_undiscounted": [
            (
                "TheNegotiatorReloaded",
                "agents.anac.y2012.TheNegotiatorReloaded.TheNegotiatorReloaded",
            ),
        ],
        "winners_welfare": [
            [
                (
                    "IAMhaggler2012",
                    "agents.anac.y2012.IAMhaggler2012.agents2011.SouthamptonAgent",
                )
            ],
            [
                (
                    "TheNegotiatorReloaded",
                    "agents.anac.y2012.TheNegotiatorReloaded.TheNegotiatorReloaded",
                )
            ],
            [("MetaAgent2012", "agents.anac.y2012.MetaAgent.MetaAgent")],
        ],
        "linear": True,
        "learning": False,
        "multilateral": False,
        "bilateral": True,
        "reservation": True,
        "discounting": True,
        "uncertainty": False,
        "elicitation": False,
    },
    2013: {
        "winners": [
            [("TheFawkes", "agents.anac.y2013.TheFawkes.TheFawkes")],
            [
                (
                    "MetaAgent",
                    "agents.anac.y2013.MetaAgent.MetaAgent2013",
                )
            ],
            [("TMFAgent", "agents.anac.y2013.TMFAgent.TMFAgent")],
        ],
        "linear": True,
        "learning": True,
        "multilateral": False,
        "bilateral": True,
        "reservation": True,
        "discounting": True,
        "uncertainty": False,
        "elicitation": False,
    },
    2014: {
        "winners": [
            [("AgentM", "agents.anac.y2014.AgentM.AgentM")],
            [("DoNA", "agents.anac.y2014.DoNA.DoNA")],
            [("Gangester", "agents.anac.y2014.Gangster.Gangster")],
        ],
        "winners_welfare": [
            [("AgentM", "agents.anac.y2014.AgentM.AgentM")],
            [("Gangester", "agents.anac.y2014.Gangster.Gangster")],
            [("E2Agent", "agents.anac.y2014.E2Agent.AnacSampleAgent")],
        ],
        "finalists": [
            [("AgentM", "agents.anac.y2014.AgentM.AgentM")],
            [("DoNA", "agents.anac.y2014.DoNA.DoNA")],
            [("Gangester", "agents.anac.y2014.Gangster.Gangster")],
            [("WhaleAgent", "agents.anac.y2014.AgentWhale.WhaleAgent")],
            [("Group2", "agents.ai2014.group2.Group2")],
            [("E2Agent", "agents.anac.y2014.E2Agent.AnacSampleAgent")],
            [("KGAgent", "agents.anac.y2014.KGAgent.KGAgent")],
            [("AgentYK", "agents.anac.y2014.AgentYK.AgentYK")],
            [("BraveCat", "agents.anac.y2014.BraveCat.BraveCat")],
        ],
        "linear": False,
        "learning": False,
        "multilateral": False,
        "bilateral": True,
        "reservation": None,
        "discounting": None,
        "uncertainty": False,
        "elicitation": False,
    },
    2015: {
        "winners": [
            [("Atlas3", "agents.anac.y2015.Atlas3.Atlas3")],
            [("ParsAgent", "agents.anac.y2015.ParsAgent.ParsAgent")],
            [("RandomDance", "agents.anac.y2015.RandomDance.RandomDance")],
        ],
        "finalists": [
            [("Atlas3", "agents.anac.y2015.Atlas3.Atlas3")],
            [("ParsAgent", "agents.anac.y2015.ParsAgent.ParsAgent")],
            [("RandomDance", "agents.anac.y2015.RandomDance.RandomDance")],
            [("Kawaii", "agents.anac.y2015.fairy.kawaii")],
            [
                (
                    "AgentBuyong",
                    "agents.anac.y2015.agentBuyogV2.AgentBuyogMain",
                )
            ],
            [("PhoenixParty", "agents.anac.y2015.Phoenix.PhoenixParty")],
            [("XianFaAgent", "agents.anac.y2015.xianfa.XianFaAgent")],
            [("PokerFace", "agents.anac.y2015.pokerface.PokerFace")],
        ],
        "linear": True,
        "learning": False,
        "multilateral": True,
        "bilateral": False,
        "reservation": None,
        "discounting": None,
        "uncertainty": False,
        "elicitation": False,
    },
    2016: {
        "winners": [
            [("Caduceus", "agents.anac.y2016.caduceus.Caduceus")],
            [
                ("YXAgent", "agents.anac.y2016.yxagent.YXAgent"),
                ("ParsCat", "agents.anac.y2016.parscat.ParsCat"),
            ],
        ],
        "finalists": [
            [("Caduceus", "agents.anac.y2016.caduceus.Caduceus")],
            [("YXAgent", "agents.anac.y2016.yxagent.YXAgent")],
            [("ParsCat", "agents.anac.y2016.parscat.ParsCat")],
            [("Farma", "agents.anac.y2017.farma.Farma17")],
            [("MyAgent", "agents.anac.y2016.myagent.MyAgent")],
            [("Atlas32016", "agents.anac.y2016.atlas3.Atlas32016")],
            [("Ngent", "agents.anac.y2016.ngent.Ngent")],
            [("GrandmaAgent", "agents.anac.y2016.grandma.GrandmaAgent")],
            [("AgentHP2", "agents.anac.y2016.agenthp2.AgentHP2_main")],
            [("Terra", "agents.anac.y2016.terra.Terra")],
        ],
        "linear": True,
        "learning": False,
        "multilateral": True,
        "bilateral": False,
        "reservation": None,
        "discounting": None,
        "uncertainty": False,
        "elicitation": False,
    },
    2017: {
        "winners": [
            [("PonPokoAgent", "agents.anac.y2017.ponpokoagent.PonPokoAgent")],
            [("CaduceusDC16", "agents.anac.y2017.caduceusdc16.CaduceusDC16")],
            [("Rubick", "agents.anac.y2017.rubick.Rubick")],
        ],
        "linear": True,
        "learning": True,
        "multilateral": True,
        "bilateral": False,
        "reservation": None,
        "discounting": None,
        "uncertainty": False,
        "elicitation": False,
    },
    2018: {
        "winners": [
            [
                (
                    "AgreeableAgent2018",
                    "agents.anac.y2018.agreeableagent2018.AgreeableAgent2018",
                )
            ],
            [("MengWan", "agents.anac.y2018.meng_wan.Agent36")],
            [("BetaOne", "agents.anac.y2018.beta_one.Group2")],
        ],
        "linear": True,
        "learning": True,
        "multilateral": True,
        "bilateral": False,
        "reservation": None,
        "discounting": True,
        "uncertainty": False,
        "elicitation": False,
    },
    # 2019: {
    #     "winners": [["AgentGG"],[ "KakeSoba"],[ "SAGA"]],
    #     "linear": True,
    #     "learning": True,
    #     "multilateral": True,
    #     "bilateral": False,
    #     "reservation": None,
    #     "discounting": None,
    #     "uncertainty": True,
    #     "elicitation": False,
    # },
    # 2020: {
    #     "winners": [],
    #     "linear": True,
    #     "learning": True,
    #     "multilateral": True,
    #     "bilateral": False,
    #     "reservation": None,
    #     "discounting": None,
    #     "uncertainty": True,
    #     "elicitation": True,
    # },
}


ALL_GENIUS_INVALID_NEGOTIATORS = [
    "agents.UIAgent",
    "agents.UIAgentExtended",
    "negotiator.parties.CounterOfferHumanNegotiationParty",
    "agents.FunctionalAcceptor",
]
"""Genius agents that are known not to work (most likely because they require human interaction)"""

ALL_GENIUS_BASIC_NEGOTIATORS = [
    "agents.ImmediateAcceptor",
    "agents.SimpleAgent",
    "negotiator.parties.RandomCounterOfferNegotiationParty",
    "negotiator.parties.RandomParty",
    "negotiator.parties.RandomParty2",
    "agents.TimeDependentAgentBoulware",
    "agents.TimeDependentAgentConceder",
    "agents.TimeDependentAgentHardliner",
    "agents.TimeDependentAgentLinear",
    "negotiator.parties.BoulwareNegotiationParty",
    "negotiator.parties.ConcederNegotiationParty",
    "agents.SimilarityAgent",
    "agents.FuzzyAgent",
    "agents.OptimalBidderSimple",
    "agents.BayesianAgent",
    "agents.ABMPAgent2",
]
"""Basic negotiators defined in Genius that are not based on any ANAC submissions"""

ALL_GENIUS_NEGOTIATORS = [
    "agents.ABMPAgent2",
    "agents.BayesianAgent",
    "agents.FunctionalAcceptor",
    "agents.FuzzyAgent",
    "agents.ImmediateAcceptor",
    "agents.OptimalBidderSimple",
    "agents.SimilarityAgent",
    "agents.SimpleAgent",
    "agents.TimeDependentAgentBoulware",
    "agents.TimeDependentAgentConceder",
    "agents.TimeDependentAgentHardliner",
    "agents.TimeDependentAgentLinear",
    "agents.UtilityBasedAcceptor",
    "negotiator.parties.BoulwareNegotiationParty",
    "negotiator.parties.ConcederNegotiationParty",
    "negotiator.parties.RandomCounterOfferNegotiationParty",
    "negotiator.parties.RandomParty",
    "negotiator.parties.RandomParty2",
    "parties.in4010.q12015.group1.Group1",
    "parties.in4010.q12015.group2.Group2",
    "parties.in4010.q12015.group3.Group3",
    "parties.in4010.q12015.group4.Group4",
    "parties.in4010.q12015.group5.Group5",
    "parties.in4010.q12015.group6.Group6",
    "parties.in4010.q12015.group7.Group7",
    "parties.in4010.q12015.group8.Group8",
    "parties.in4010.q12015.group9.Group9",
    "parties.in4010.q12015.group10.Group10",
    "parties.in4010.q12015.group11.Group11",
    "parties.in4010.q12015.group12.Group12",
    "parties.in4010.q12015.group13.Group13",
    "parties.in4010.q12015.group14.Group14",
    "parties.in4010.q12015.group15.Group15",
    "parties.in4010.q12015.group16.Group16",
    "parties.in4010.q12015.group17.Group17",
    "parties.in4010.q12015.group18.Group18",
    "parties.in4010.q12015.group19.Group19",
    "parties.in4010.q12015.group20.Group20",
    "parties.in4010.q12015.group21.Group21",
    "parties.in4010.q12015.group22.Group22",
    "agents.anac.y2010.AgentFSEGA.AgentFSEGA",
    "agents.anac.y2010.AgentK.Agent_K",
    "agents.anac.y2010.AgentSmith.AgentSmith",
    "agents.anac.y2010.Nozomi.Nozomi",
    "agents.anac.y2010.Southampton.IAMcrazyHaggler",
    "agents.anac.y2010.Southampton.IAMhaggler",
    "agents.anac.y2010.Yushu.Yushu",
    "agents.anac.y2011.AgentK2.Agent_K2",
    "agents.anac.y2011.BramAgent.BRAMAgent",
    "agents.anac.y2011.Gahboninho.Gahboninho",
    "agents.anac.y2011.HardHeaded.KLH",
    "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011",
    "agents.anac.y2011.Nice_Tit_for_Tat.NiceTitForTat",
    "agents.anac.y2011.TheNegotiator.TheNegotiator",
    "agents.anac.y2011.ValueModelAgent.ValueModelAgent",
    "agents.anac.y2012.AgentLG.AgentLG",
    "agents.anac.y2012.AgentMR.AgentMR",
    "agents.anac.y2012.BRAMAgent2.BRAMAgent2",
    "agents.anac.y2012.CUHKAgent.CUHKAgent",
    "agents.anac.y2012.IAMhaggler2012.IAMhaggler2012",
    "agents.anac.y2012.MetaAgent.MetaAgent",
    "agents.anac.y2012.OMACagent.OMACagent",
    "agents.anac.y2012.TheNegotiatorReloaded.TheNegotiatorReloaded",
    "agents.anac.y2013.AgentKF.AgentKF",
    "agents.anac.y2013.GAgent.AgentI",
    "agents.anac.y2013.InoxAgent.InoxAgent",
    "agents.anac.y2013.MetaAgent.MetaAgent2013",
    "agents.anac.y2013.SlavaAgent.SlavaAgent",
    "agents.anac.y2013.TMFAgent.TMFAgent",
    "agents.anac.y2013.TheFawkes.TheFawkes",
    "agents.anac.y2014.AgentM.AgentM",
    "agents.anac.y2014.AgentQuest.AgentQuest",
    "agents.anac.y2014.AgentTD.AgentTD",
    "agents.anac.y2014.AgentTRP.AgentTRP",
    "agents.anac.y2014.AgentWhale.WhaleAgent",
    "agents.anac.y2014.AgentYK.AgentYK",
    "agents.anac.y2014.ArisawaYaki.ArisawaYaki",
    "agents.anac.y2014.Aster.Aster",
    "agents.anac.y2014.Atlas.Atlas",
    "agents.anac.y2014.BraveCat.BraveCat",
    "agents.anac.y2014.DoNA.DoNA",
    "agents.anac.y2014.E2Agent.AnacSampleAgent",
    "agents.anac.y2014.Flinch.Flinch",
    "agents.anac.y2014.Gangster.Gangster",
    "agents.anac.y2014.KGAgent.KGAgent",
    "agents.anac.y2014.SimpaticoAgent.Simpatico",
    "agents.anac.y2014.Sobut.Sobut",
    "agents.anac.y2014.TUDelftGroup2.Group2Agent",
    "agents.anac.y2015.AgentHP.AgentHP",
    "agents.anac.y2015.AgentNeo.Groupn",
    "agents.anac.y2015.AgentW.AgentW",
    "agents.anac.y2015.AgentX.AgentX",
    "agents.anac.y2015.AresParty.AresParty",
    "agents.anac.y2015.Atlas3.Atlas3",
    "agents.anac.y2015.DrageKnight.DrageKnight",
    "agents.anac.y2015.JonnyBlack.JonnyBlack",
    "agents.anac.y2015.Mercury.Mercury",
    "agents.anac.y2015.ParsAgent.ParsAgent",
    "agents.anac.y2015.Phoenix.PhoenixParty",
    "agents.anac.y2015.RandomDance.RandomDance",
    "agents.anac.y2015.SENGOKU.SENGOKU",
    "agents.anac.y2015.TUDMixedStrategyAgent.TUDMixedStrategyAgent",
    "agents.anac.y2015.agentBuyogV2.AgentBuyogMain",
    "agents.anac.y2015.agenth.AgentH",
    "agents.anac.y2015.cuhkagent2015.CUHKAgent2015",
    "agents.anac.y2015.fairy.kawaii",
    "agents.anac.y2015.group2.Group2",
    "agents.anac.y2015.meanBot.MeanBot",
    "agents.anac.y2015.pnegotiator.PNegotiator",
    "agents.anac.y2015.pokerface.PokerFace",
    "agents.anac.y2015.xianfa.XianFaAgent",
    "agents.anac.y2016.agenthp2.AgentHP2_main",
    "agents.anac.y2016.agentlight.AgentLight",
    "agents.anac.y2016.agentsmith.AgentSmith2016",
    "agents.anac.y2016.atlas3.Atlas32016",
    "agents.anac.y2016.caduceus.Caduceus",
    "agents.anac.y2016.clockworkagent.ClockworkAgent",
    "agents.anac.y2016.farma.Farma",
    "agents.anac.y2016.grandma.GrandmaAgent",
    "agents.anac.y2016.maxoops.MaxOops",
    "agents.anac.y2016.myagent.MyAgent",
    "agents.anac.y2016.ngent.Ngent",
    "agents.anac.y2016.pars2.ParsAgent2",
    "agents.anac.y2016.parscat.ParsCat",
    "agents.anac.y2016.syagent.SYAgent",
    "agents.anac.y2016.terra.Terra",
    "agents.anac.y2016.yxagent.YXAgent",
    "agents.anac.y2017.agentf.AgentF",
    "agents.anac.y2017.agentkn.AgentKN",
    "agents.anac.y2017.caduceusdc16.CaduceusDC16",
    "agents.anac.y2017.mamenchis.Mamenchis",
    "agents.anac.y2017.parsagent3.ShahAgent",
    "agents.anac.y2017.ponpokoagent.PonPokoAgent",
    "agents.anac.y2017.rubick.Rubick",
    "agents.anac.y2017.simpleagent.SimpleAgent",
    "agents.anac.y2018.agent33.Agent33",
    "agents.anac.y2018.agentherb.AgentHerb",
    "agents.anac.y2018.agentnp1.AgentNP1",
    "agents.anac.y2018.agreeableagent2018.AgreeableAgent2018",
    "agents.anac.y2018.ateamagent.ATeamAgent",
    "agents.anac.y2018.beta_one.Group2",
    "agents.anac.y2018.condagent.ConDAgent",
    "agents.anac.y2018.exp_rubick.Exp_Rubick",
    "agents.anac.y2018.fullagent.FullAgent",
    "agents.anac.y2018.groupy.GroupY",
    "agents.anac.y2018.iqson.IQSun2018",
    "agents.anac.y2018.lancelot.Lancelot",
    "agents.anac.y2018.libra.Libra",
    "agents.anac.y2018.meng_wan.Agent36",
    "agents.anac.y2018.ponpokorampage.PonPokoRampage",
    "agents.anac.y2018.seto.Seto",
    "agents.anac.y2018.shiboy.Shiboy",
    "agents.anac.y2018.smac_agent.SMAC_Agent",
    "agents.anac.y2018.sontag.Sontag",
    "agents.anac.y2018.yeela.Yeela",
]
"""All Negotiators Accessible through Genius UI"""


ALL_NON_SOAP_GENIUS_NEGOTIATORS = [
    "parties.AlternatingMultipleOffers.RandomAmopParty",
    "parties.AlternatingMultipleOffers.RandomAmopPartyMajority",
    "parties.simplemediator.RandomFlippingMediator",
    "parties.simplemediator.FixedOrderFlippingMediator",
    "parties.simplemediator.HillClimber",
    "parties.simplemediator.Annealer",
    "parties.simplemediator.FeedbackParty",
    "parties.feedbackmediator.FeedbackMediator",
    "parties.FeedbackHillClimber",
]
"""All Genius negotiators for protocols other than SAOP"""

ALL_GENIUS_AMOP_NEGOTIATORS = [
    "parties.AlternatingMultipleOffers.RandomAmopParty",
    "parties.AlternatingMultipleOffers.RandomAmopPartyMajority",
]
"""All Genius negotiators for the Alternating Multiple Offers Protocol"""

ALL_GENIUS_SIMPLE_MEDIATOR_NEGOTIATORS = [
    "parties.simplemediator.RandomFlippingMediator",
    "parties.simplemediator.FixedOrderFlippingMediator",
    "parties.simplemediator.HillClimber",
    "parties.simplemediator.Annealer",
    "parties.simplemediator.FeedbackParty",
]
"""All Genius negotiators for the Simple Mediator Protocol"""

ALL_GENIUS_FEEDBACK_MEDIATOR_NEGOTIATORS = [
    "parties.feedbackmediator.FeedbackMediator",
]
"""All Genius negotiators for the Feedback Mediator Protocol"""

AGENT_BASED_NEGOTIATORS = [
    "agents.ABMPAgent",
    "agents.ABMPAgent2",
    "agents.BayesianAgent",
    "agents.BayesianAgentForAuction",
    "agents.DecUtilAgent",
    "agents.FuzzyAgent",
    "agents.OptimalBidder",
    "agents.OptimalBidderU",
    "agents.QOAgent",
    "agents.RandomIncreasingUtilAgent",
    "agents.SimilarityAgent",
    "agents.anac.y2010.AgentFSEGA.AgentFSEGA",
    "agents.anac.y2010.AgentK.Agent_K",
    "agents.anac.y2010.AgentSmith.AgentSmith",
    "agents.anac.y2010.Nozomi.Nozomi",
    "agents.anac.y2010.Southampton.SouthamptonAgent",
    "agents.anac.y2010.Yushu.Yushu",
    "agents.anac.y2011.AgentK2.Agent_K2",
    "agents.anac.y2011.BramAgent.BRAMAgent",
    "agents.anac.y2011.Gahboninho.Gahboninho",
    "agents.anac.y2011.HardHeaded.KLH",
    "agents.anac.y2011.IAMhaggler2011.IAMhaggler2011",
    "agents.anac.y2011.Nice_Tit_for_Tat.NiceTitForTat",
    "agents.anac.y2011.TheNegotiator.TheNegotiator",
    "agents.anac.y2011.ValueModelAgent.ValueModelAgent",
    "agents.anac.y2012.AgentLG.AgentLG",
    "agents.anac.y2012.TheNegotiatorReloaded.TheNegotiatorReloaded",
    "agents.anac.y2012.AgentMR.AgentMR",
    "agents.anac.y2012.BRAMAgent2.BRAMAgent2",
    "agents.anac.y2012.CUHKAgent.CUHKAgent",
    "agents.anac.y2012.MetaAgent.MetaAgent",
    "agents.anac.y2012.OMACagent.OMACagent",
    "agents.anac.y2013.AgentKF.AgentKF",
    "agents.anac.y2013.GAgent.AgentI",
    "agents.anac.y2013.MetaAgent.MetaAgent2013",
    "agents.anac.y2013.SlavaAgent.SlavaAgent",
    "agents.anac.y2013.TMFAgent.TMFAgent",
    "agents.anac.y2013.TheFawkes.TheFawkes",
    "agents.anac.y2014.AgentM.AgentM",
    "agents.anac.y2014.AgentQuest.AgentQuest",
    "agents.anac.y2014.AgentTD.AgentTD",
    "agents.anac.y2014.AgentTRP.AgentTRP",
    "agents.anac.y2014.AgentWhale.WhaleAgent",
    "agents.anac.y2014.AgentYK.AgentYK",
    "agents.anac.y2014.ArisawaYaki.ArisawaYaki",
    "agents.anac.y2014.Aster.Aster",
    "agents.anac.y2014.Atlas.Atlas",
    "agents.anac.y2014.BraveCat.BraveCat",
    "agents.anac.y2014.DoNA.DoNA",
    "agents.anac.y2014.E2Agent.AnacSampleAgent",
    "agents.anac.y2014.Flinch.Flinch",
    "agents.anac.y2014.Gangster.Gangster",
    "agents.anac.y2014.KGAgent.KGAgent",
    "agents.anac.y2014.SimpaticoAgent.Simpatico",
    "agents.anac.y2014.Sobut.Sobut",
    "agents.anac.y2017.geneking.GeneKing",
]
"""Genius agents based on the Agent base-class. These are the oldest agents"""

PARTY_BASED_NEGOTIATORS = [
    "agents.ai2014.group1.Group1",
    "agents.ai2014.group10.Group10",
    "agents.ai2014.group11.Group11",
    "agents.ai2014.group12.Group12",
    "agents.ai2014.group2.Group2",
    "agents.ai2014.group3.Group3",
    "agents.ai2014.group4.Group4",
    "agents.ai2014.group5.Group5",
    "agents.ai2014.group6.Group6",
    "agents.ai2014.group7.Group7",
    "agents.ai2014.group9.Group9",
    "agents.anac.y2015.AgentNeo.Groupn",
    "agents.anac.y2015.AgentX.AgentX",
    "agents.anac.y2015.AresParty.AresParty",
    "agents.anac.y2015.Atlas3.Atlas3",
    "agents.anac.y2015.DrageKnight.DrageKnight",
    "agents.anac.y2015.JonnyBlack.JonnyBlack",
    "agents.anac.y2015.Mercury.Mercury",
    "agents.anac.y2015.ParsAgent.ParsAgent",
    "agents.anac.y2015.Phoenix.PhoenixParty",
    "agents.anac.y2015.RandomDance.RandomDance",
    "agents.anac.y2015.SENGOKU.SENGOKU",
    "agents.anac.y2015.TUDMixedStrategyAgent.TUDMixedStrategyAgent",
    "agents.anac.y2015.agentBuyogV2.AgentBuyogMain",
    "agents.anac.y2015.agenth.AgentH",
    "agents.anac.y2015.cuhkagent2015.CUHKAgent2015",
    "agents.anac.y2015.fairy.kawaii",
    "agents.anac.y2015.group2.Group2",
    "agents.anac.y2015.meanBot.MeanBot",
    "agents.anac.y2015.pokerface.PokerFace",
    "agents.anac.y2016.agentsmith.AgentSmith2016",
    "agents.anac.y2016.atlas3.Atlas32016",
    "agents.anac.y2016.caduceus.Caduceus",
    "agents.anac.y2016.clockworkagent.ClockworkAgent",
    "agents.anac.y2016.farma.Farma",
    "agents.anac.y2016.grandma.GrandmaAgent",
    "agents.anac.y2016.myagent.MyAgent",
    "agents.anac.y2016.ngent.Ngent",
    "agents.anac.y2016.pars2.ParsAgent2",
    "agents.anac.y2016.parscat.ParsCat",
    "agents.anac.y2016.syagent.SYAgent",
    "agents.anac.y2016.terra.Terra",
    "agents.anac.y2016.yxagent.YXAgent",
    "agents.anac.y2017.agentf.AgentF",
    "agents.anac.y2017.caduceusdc16.CaduceusDC16",
    "agents.anac.y2017.farma.Farma17",
    "agents.anac.y2017.ponpokoagent.PonPokoAgent",
    "agents.anac.y2017.tangxun.taxibox",
    "agents.anac.y2017.tucagent.TucAgent",
]
"""Genius agents based on the Party base-class. These are the newest agents"""

TEST_FAILING_NEGOTIATORS = [
    "agents.ABMPAgent2",  # failed some but not all tests
    "agents.BayesianAgentForAuction",
    "agents.DecUtilAgent",
    "agents.FuzzyAgent",  # fails most test but not all
    "agents.OptimalBidder",
    "agents.OptimalBidderSimple",  # failes most tests but not all
    "agents.OptimalBidderU",
    "agents.QOAgent",
    "agents.RandomIncreasingUtilAgent",
    "agents.SimilarityAgent",  # failed some but not all tests
    "agents.anac.y2010.Southampton.SouthamptonAgent",
    "agents.anac.y2013.MetaAgent.MetaAgent2013",  # failed a few tests
    "agents.anac.y2014.AgentTD.AgentTD",
    "agents.anac.y2014.ArisawaYaki.ArisawaYaki",
    "agents.anac.y2014.Aster.Aster",
    "agents.anac.y2014.Atlas.Atlas",
    "agents.anac.y2014.E2Agent.AnacSampleAgent",
    "agents.anac.y2014.KGAgent.KGAgent",
    "agents.anac.y2014.TUDelftGroup2.Group2Agent",
    "agents.anac.y2014.simpleagent.SimpleAgent",
    "agents.anac.y2015.AgentHP.AgentHP",
    "agents.anac.y2015.AgentW.AgentW",
    "agents.anac.y2016.agenthp2.AgentHP2_main",
    "agents.anac.y2016.agentlight.SgentLight",
    "agents.anac.y2016.maxoops.MaxOops",
    "agents.anac.y2016.ngent.Ngent",
    "agents.anac.y2017.agentkn.AgentKN",
    "agents.anac.y2017.caduceusdc16.CaduceusDC16",  # failed all tests
    "agents.anac.y2017.geneking.GeneKing",
    "agents.anac.y2017.mamenchis.Mamenchis",
    "agents.anac.y2017.parsagent3.ShahAgent",
    "agents.anac.y2017.rubick.Rubick",  # failed all tests
    "parties.in4010.q12015.group1.Group1",
    "parties.in4010.q12015.group10.Group10",
    "parties.in4010.q12015.group13.Group13",
    "parties.in4010.q12015.group18.Group18",
    "parties.in4010.q12015.group21.Group21",
    "parties.in4010.q12015.group7.Group7",
    "agents.anac.y2015.cuhkagent2015.CUHKAgent2015",
    "agents.anac.y2017.simpleagent.SimpleAgent",
    "agents.ai2014.group4.Group4",
    "parties.in4010.q12015.group6.Group6",
    "agents.anac.y2016.agentlight.AgentLight",
    "parties.in4010.q12015.group3.Group3",
    "parties.in4010.q12015.group19.Group19",
    "agents.anac.y2015.Phoenix.PhoenixParty",
    "agents.anac.y2015.Mercury.Mercury",
    "agents.anac.y2015.xianfa.XianFaAgent",
    "agents.anac.y2016.grandma.GrandmaAgent",
    "agents.anac.y2015.pnegotiator.PNegotiator",
] + [_ for _ in ALL_GENIUS_NEGOTIATORS if "y2018" in _]

"""Agetns taht fail simple tests making them less robust over the bridge"""

TESTED_NEGOTIATORS = list(
    set(
        ["agents.anac.y2015.AgentX.AgentX"]
        + list(
            itertools.chain(
                *(
                    list(_[1] for _ in itertools.chain(*v["winners"]))
                    for year, v in GENIUS_INFO.items()
                    if v["multilateral"] and not v["learning"]
                )
            )
        )
    )
    - set(TEST_FAILING_NEGOTIATORS)
)
"""Some of the most tested negotaitors"""

ALL_NEGOTIATORS = list(
    set(PARTY_BASED_NEGOTIATORS + AGENT_BASED_NEGOTIATORS + ALL_GENIUS_NEGOTIATORS)
)
"""All Genius Negotiators accessible from NegMAS"""

ALL_PASSING_NEGOTIATORS = list(
    set(ALL_NEGOTIATORS)
    - set(TEST_FAILING_NEGOTIATORS)
    - set(ALL_GENIUS_INVALID_NEGOTIATORS)
)
"""All negotiators that passed simple tests showing they work on the bridge"""

ALL_ANAC_AGENTS = [_ for _ in ALL_NEGOTIATORS if "anac" in _]
"""All agents submitted to ANAC and registered in Genius"""


def get_name(java_class: str) -> str:
    """Returns the name of the agent with the given class"""
    return java_class.split(".")[-1]


def get_java_class(name) -> str | None:
    """Returns the java class for the agent with this name if known otherwise it returns None"""
    for class_name in ALL_NEGOTIATORS:
        if name in class_name:
            return class_name
    return None


def get_anac_agents(
    *,
    year: int | None = None,
    linear: bool | None = None,
    learning: bool | None = None,
    multilateral: bool | None = None,
    bilateral: bool | None = None,
    reservation: bool | None = None,
    discounting: bool | None = None,
    uncertainty: bool | None = None,
    elicitation: bool | None = None,
    winners_only: bool = False,
    finalists_only: bool = False,
) -> list[tuple[str, str]]:
    """
    Get Genius agents matching some given criteria

    Returns:
        a list of 2-item tuples giving agent name and java class name.

    Remarks:
        - For all criteria other than winners_only, and finalists_only, passing
          None means no constraints otherwise the game on the given year must
          match the criterion value (True or False) to get agents from this year.
        - This function uses a heuristic and may not be completely accurate. Use
          with caution.
    """

    def get_agents(year, d) -> set[tuple[str, str]]:

        if winners_only:
            lst = d.get("winners", [[]])
            lst = list(itertools.chain(*lst))
        elif finalists_only:
            lst = d.get("finalists", [])
            lst = list(itertools.chain(*lst))
        else:
            lst = [_ for _ in ALL_NEGOTIATORS if str(year) in _]
            lst = [(_.split(".")[-1], _) for _ in lst]
        # lst = tuple(lst)
        # if winners_only or finalists_only:
        #     return set(itertools.chain(*lst))
        return set(lst)

    agents: set[tuple[str, str]] = set()
    if year is None:
        for y in GENIUS_INFO.keys():
            agents = agents.union(
                get_anac_agents(
                    year=y,
                    linear=linear,
                    learning=learning,
                    multilateral=multilateral,
                    bilateral=bilateral,
                    reservation=reservation,
                    discounting=discounting,
                    uncertainty=uncertainty,
                    elicitation=elicitation,
                    winners_only=winners_only,
                    finalists_only=finalists_only,
                )
            )
        return list(agents)

    d = GENIUS_INFO.get(year, None)
    if not d:
        return []
    if linear is None or (linear is not None and d["linear"] == linear):
        agents = get_agents(year, d)
    if learning is None or (learning is not None and d["learning"] == learning):
        agents = agents.intersection(get_agents(year, d))
    if bilateral is None or (bilateral is not None and d["bilateral"] == bilateral):
        agents = agents.intersection(get_agents(year, d))
    if multilateral is None or (
        multilateral is not None and d["multilateral"] == multilateral
    ):
        agents = agents.intersection(get_agents(year, d))
    if discounting is None or (
        discounting is not None and d["discounting"] == discounting
    ):
        agents = agents.intersection(get_agents(year, d))
    if reservation is None or (
        reservation is not None and d["reservation"] == reservation
    ):
        agents = agents.intersection(get_agents(year, d))
    if uncertainty is None or (
        uncertainty is not None and d["uncertainty"] == uncertainty
    ):
        agents = agents.intersection(get_agents(year, d))
    if elicitation is None or (
        elicitation is not None and d["elicitation"] == elicitation
    ):
        agents = agents.intersection(get_agents(year, d))
    return list(agents)



"""
Implements GeniusBridge which manages connections to Genius through Py4J.

The main class is the GeniusBridge class which encapsulates a gateway
(connection) to a JVM running the geniusbridge.jar file.
We can have multiple gateways connected to the same JVM bridge.

The bridge life-cycle is as follows:
start -> connect -> stop/restart

The most important methods that this class provides are:

Status enquiry
--------------
- is_running: Tells you if a bridge is running on a given port
- is_installed: Tells you if the bridge jar is insalled in the default location

Bridge Lifetime Control
-----------------------
- start() starts a bridge and connects to it
- connect() connects to a *running* bridge
- stop() stops a running bridge.
- restart() stops then starts a bridge.

Bridge Control Operations
-------------------------

- clean() removes all agents from the bridge and runs garbage collection. You
  must be sure that no active negotiations are happening when you call this
  method.
- kill_threads() [not recommended] Kills all threads started by the bridge. The
  bridge is most likely going to become unusable after running this command


"""
from __future__ import annotations

import os
import pathlib
import socket
import subprocess
import time
import traceback
from pathlib import Path
from typing import Any

import psutil
from py4j.java_gateway import (
    CallbackServerParameters,
    GatewayParameters,
    JavaGateway,
    JavaObject,
)
from py4j.protocol import Py4JNetworkError

import negmas.warnings as warnings

from ..config import CONFIG_KEY_GENIUS_BRIDGE_JAR, NEGMAS_CONFIG
from ..helpers import TimeoutCaller, TimeoutError, unique_name
from .common import DEFAULT_JAVA_PORT, get_free_tcp_port

__all__ = [
    "GeniusBridge",
    "init_genius_bridge",
    "genius_bridge_is_running",
    "genius_bridge_is_installed",
]


def _kill_process(proc_pid):
    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()


def genius_bridge_is_running(port: int = DEFAULT_JAVA_PORT) -> bool:
    """
    Checks whether a Genius Bridge is running. A genius bridge allows you to use `GeniusNegotiator` objects.

    Remarks:

        You can start a Genius Bridge in at least three ways:

        - execute the python function `init_genius_bridge()` in this module
        - run "negmas genius" on the terminal
        - execute `GeniusBridge.start()`

    """
    s = socket.socket()
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        s.connect(("127.0.0.1", port))
        s.close()
        # we know someone is listening. Now we check that it is us
        gateway = JavaGateway(
            gateway_parameters=GatewayParameters(port=port, auto_close=True),
            callback_server_parameters=CallbackServerParameters(
                port=0, daemonize=True, daemonize_connections=True
            ),
        )
        gateway.jvm.System.currentTimeMillis()  # type: ignore
        return True
    except ConnectionRefusedError:
        # try:
        #     s.shutdown(2)
        # except:
        #     pass
        s.close()
        return False
    except IndexError:
        # try:
        #     s.shutdown(2)
        # except:
        #     pass
        s.close()
        return False
    except Py4JNetworkError:
        # try:
        #     s.shutdown(2)
        # except:
        #     pass
        s.close()
        return False


def init_genius_bridge(
    path: Path | str | None = None,
    port: int = DEFAULT_JAVA_PORT,
    debug: bool = False,
    timeout: float = 0,
) -> bool:
    """Initializes a genius connection

    Args:
        path: The path to a JAR file that runs negloader
        port: port number to use
        debug: If true, passes --debug to the bridge
        timeout: If positive and nonzero, passes it as the global timeout for the bridge. Note that
                 currently, the bridge supports only integer timeout values and the fraction will be
                 truncated.

    Returns:
        True if successful

    """
    if port <= 0:
        port = get_free_tcp_port()
    if genius_bridge_is_running(port):
        return True
    # if not force and common_gateway is not None and common_port == port:
    #     print("Java already initialized")
    #     return True

    if not path:
        path = NEGMAS_CONFIG.get(CONFIG_KEY_GENIUS_BRIDGE_JAR, None)
    if path is None:
        warnings.warn(
            "Cannot find the path to genius bridge jar. Download the jar somewhere in your machine and add its path"
            'to ~/negmas/config.json under the key "genius_bridge_jar".\n\nFor example, if you downloaded the jar'
            " to /path/to/your/jar then edit ~/negmas/config.json to read something like\n\n"
            '{\n\t"genius_bridge_jar": "/path/to/your/jar",\n\t.... rest of the config\n}\n\n'
            "You can find the jar at http://www.yasserm.com/scml/genius-8.0.4-bridge.jar",
            warnings.NegmasBridgePathWarning,
        )
        return False
    path = Path(path).expanduser().absolute()
    if debug:
        params = " --debug"
    else:
        params = " --silent --no-logs"
    if timeout >= 0:
        params += f" --timeout={int(timeout)}"

    try:
        subprocess.Popen(  # ['java', '-jar',  path, '--die-on-exit', f'{port}']
            f"java -jar {path} --die-on-exit {params} {port}", shell=True
        )
    except (OSError, TimeoutError, RuntimeError, ValueError):
        return False
    time.sleep(0.5)
    gateway = JavaGateway(
        gateway_parameters=GatewayParameters(port=port, auto_close=True),
        callback_server_parameters=CallbackServerParameters(
            port=0, daemonize=True, daemonize_connections=True
        ),
    )
    python_port = gateway.get_callback_server().get_listening_port()  # type: ignore
    gateway.java_gateway_server.resetCallbackClient(  # type: ignore
        gateway.java_gateway_server.getCallbackClient().getAddress(), python_port  # type: ignore
    )
    return True


def genius_bridge_is_installed() -> bool:
    """
    Checks if geniusbridge is available in the default path location
    """
    return (Path.home() / "negmas" / "files" / "geniusbridge.jar").exists()


class GeniusBridge:
    gateways: dict[int, GeniusBridge] = dict()
    java_processes: dict[int, Any] = dict()
    python_ports: dict[int, int] = dict()

    def __init__(
        self,
    ):
        raise RuntimeError(f"Cannot create objects of type GeniusBridge.")

    """Gateways to different genius bridges"""
    """handles to the java processes running the bridges"""
    """The port used by python's Gateway to connect to the bridge"""

    @classmethod
    def is_running(cls, port: int) -> bool:
        """Returns true if a geniusbridge.jar is running on the given port"""
        return genius_bridge_is_running(port)

    @classmethod
    def is_installed(cls) -> bool:
        """Returns true if a geniusbridge.jar is available"""
        return genius_bridge_is_installed()

    @classmethod
    def gateway(cls, port=DEFAULT_JAVA_PORT, force=False):
        """
        Finds and returns a gateway for a genius bridge on the given port

        Args:
            port: The port used by the Jave genius bridge.
            force: If true, a new gateway is created even if one exists in
                   the list of gateways available in `GeniusBridge`.gateways.
        Returns:
            The gateway if found otherwise an exception will be thrown

        Remarks:
            - this method does NOT start a bridge. It only connects to a
              running bridge.
        """
        assert port > 0
        gateway = cls.gateways.get(port, None) if not force else None
        if gateway is not None:
            return gateway
        try:
            gateway = JavaGateway(
                gateway_parameters=GatewayParameters(port=port, auto_close=True),
                callback_server_parameters=CallbackServerParameters(
                    port=0, daemonize=True, daemonize_connections=True
                ),
            )
            python_port = gateway.get_callback_server().get_listening_port()  # type: ignore
            gateway.java_gateway_server.resetCallbackClient(  # type: ignore
                gateway.java_gateway_server.getCallbackClient().getAddress(),  # type: ignore
                python_port,
            )
        except:
            if gateway is not None:
                gateway.shutdown()
                gateway.shutdown_callback_server()
            return None
        cls.python_ports[port] = python_port
        cls.gateways[port] = gateway  # type: ignore
        return gateway

    @classmethod
    def wait_until_listening(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        timeout: float = 0.5,
    ) -> bool:
        """
        waits until a genius bridge is  listening to the given port

        Args:
            port: The port to test
            timeout: Maximum time to wait before returning (in seconds)

        Returns:
            True if the genius bridge is running any more (success).
        """
        if genius_bridge_is_running(port):
            return True
        time.sleep(timeout)
        return TimeoutCaller.run(
            lambda: genius_bridge_is_running(port), timeout=timeout
        )

    @classmethod
    def start(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        path: str | None = None,
        debug: bool = False,
        timeout: float = 0,
        force_timeout: bool = True,
        save_logs: bool = False,
        log_path: os.PathLike | None = None,
        die_on_exit: bool = False,
        use_shell: bool = False,
    ) -> int:
        """Initializes a genius connection

        Args:
            port: port number to use. A value <= 0 means get any free tcp port.
            path: The path to a JAR file that runs negloader
            debug: If true, passes --debug to the bridge
            timeout: If positive and nonzero, passes it as the global timeout for the bridge. Note that
                     currently, the bridge supports only integer timeout values and the fraction will be
                     truncated.
            force_timeout: if false, no timeout will be forced by the bridge
            save_logs: If false, the brige is instructed not to save any logs
            log_path: the path to store logs from the bridge. Onle effective if `save_logs`
                     If not given, defaults to ~/negmas/geniusbridge/logs/{port}-{datetime}.txt
            die_on_exit: If given, the bridge will be closed when this process is ended
            use_shell: If given, the bridge will be started in  a subshell.

        Returns:
           The port number used by the java process. 0 for failure

        Remarks:
            - if a bridge is running, it will return its port and it does not matter whether
              or not the bridge is started from this process or any other way.
            - it is recommended not to change the defaults for this function.

        """
        if port <= 0:
            port = get_free_tcp_port()
        if cls.is_running(port):
            return 0 if cls.gateway(port) is None else port
        path = (
            NEGMAS_CONFIG.get(CONFIG_KEY_GENIUS_BRIDGE_JAR, None)
            if path is None or not path
            else path
        )
        if path is None:
            warnings.warn(
                "Cannot find the path to genius bridge jar. Download the jar somewhere in your machine and add its path"
                'to ~/negmas/config.json under the key "genius_bridge_jar".\n\nFor example, if you downloaded the jar'
                " to /path/to/your/jar then edit ~/negmas/config.json to read something like\n\n"
                '{\n\t"genius_bridge_jar": "/path/to/your/jar",\n\t.... rest of the config\n}\n\n'
                "You can find the jar at http://www.yasserm.com/scml/geniusbridge.jar",
                warnings.NegmasBridgePathWarning,
            )
            return 0
        path = Path(path).expanduser().absolute()  # type: ignore
        if log_path is None or not log_path:
            log_path = (
                Path.home()
                / "negmas"
                / "geniusbridge"
                / "logs"
                / f"{unique_name(str(port), add_time=True, rand_digits=4, sep='.')}.txt"
            ).absolute()
            log_path.parent.mkdir(parents=True, exist_ok=True)
        else:
            log_path = Path(log_path).absolute()
            log_path.parent.mkdir(parents=True, exist_ok=True)
        # if die_on_exit:
        #     if debug:
        #         params = " --die-on-exit --debug"
        #     else:
        #         params = " --die-on-exit"
        #     if force_timeout:
        #         params += " --force-timeout"
        #         if timeout >= 0:
        #             params += f" --timeout={int(timeout)}"
        #     else:
        #         params += " --no-timeout"
        #     params += " --with-logs" if save_logs else " --no-logs"
        #     if save_logs:
        #         params += f"--log-file={str(log_path)}"
        # else:
        if debug:
            params = ["--debug"]
        else:
            params = ["--silent", "--no-logs"]
        if die_on_exit:
            params.append("--die-on-exit")
        if force_timeout:
            params.append("--force-timeout")
            if timeout >= 0:
                params.append(f"--timeout={int(timeout)}")
        else:
            params.append("--no-timeout")
        params.append("--with-logs" if save_logs else "--no-logs")
        if save_logs:
            params.append(f"--logfile={str(log_path)}")

        try:
            # if die_on_exit:
            #     cls.java_processes[port] = subprocess.Popen(
            #         f"java -jar {str(path)} {params} {port}",
            #         shell=use_shell,
            #         cwd=path.parent,
            #     )
            # else:
            cls.java_processes[port] = subprocess.Popen(
                ["java", "-jar", str(path)] + params + [f"{port}"],
                shell=use_shell,
                cwd=path.parent,  # type: ignore
            )
        except (OSError, TimeoutError, RuntimeError, ValueError) as e:
            warnings.warn(str(e), warnings.NegmasBridgeProcessWarning)
            return 0
        cls.wait_until_listening(port, timeout=0.1)
        return port if cls.gateway(port, force=True) is not None else 0

    @classmethod
    def _close_gateway(cls, port):
        """
        Closes the gateway and removes it from all class dicts.
        """
        gateway = cls.gateways.get(port, None)
        if gateway is None:
            cls.java_processes.pop(port, None)
            cls.python_ports.pop(port, None)

        gateway.shutdown()  # type: ignore
        gateway.shutdown_callback_server()  # type: ignore
        cls.gateways.pop(port, None)
        cls.java_processes.pop(port, None)
        cls.python_ports.pop(port, None)

    @classmethod
    def close_gateway(cls, port=DEFAULT_JAVA_PORT):
        """
        Closes the gateway.

        Args:
            port: The port the gateway is connected to.
                  If None, DEFAULT_JAVA_PORT is used.
        """
        if port is None:
            port = DEFAULT_JAVA_PORT
        cls._close_gateway(port)

    @classmethod
    def close_gateways(cls):
        """
        Closes all open gateways.
        """
        keys = list(cls.gateways.keys())
        for p in keys:
            cls._close_gateway(p)

    @classmethod
    def wait_until_not_listening(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        timeout: float = 0.5,
    ) -> bool:
        """
        waits until the genius bridge is not listening to the given port

        Args:
            port: The port to test
            max_sleep: Maximum time to wait before returning (in seconds)

        Returns:
            True if the genius bridge is NOT running any more (success).
        """
        if not genius_bridge_is_running(port):
            return False
        time.sleep(timeout)
        return not genius_bridge_is_running(port)

    @classmethod
    def shutdown(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        wait: bool = True,
    ) -> bool:
        """
        Attempts to shutdown the bridge on that port.

        Args:
            port: The port to shutdown.

        Remarks:
            - This is the cleanest way to close a java bridge and it simply sends a
              message to the bridge to shut itself down and cleanly shuts down the
              py4j bridge.
        """
        assert port > 0
        try:
            gateway = cls.gateway(port)
        except:
            return not genius_bridge_is_running(port)
        if gateway is None:
            return True
        try:
            gateway.entry_point.shutdown()
        except:
            pass
        if wait:
            cls.wait_until_not_listening(port)
        cls._close_gateway(port)
        return True

    @classmethod
    def kill(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        wait: bool = True,
    ) -> bool:
        """Kills the java bridge connected to this port by asking it to exit"""
        assert port > 0
        try:
            gateway = cls.gateway(port, force=True)
        except:
            return False
        if gateway is None:
            return False
        try:
            gateway.entry_point.kill()
        except:
            pass
        if wait:
            cls.wait_until_not_listening(port)
        cls._close_gateway(port)
        return True

    @classmethod
    def kill_forced(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        wait: bool = True,
    ) -> bool:
        """Kills the java bridge connected to this port forcibly.

        Remarks:
            - The java bridge process must have been started by this process.
        """
        assert port > 0
        p = cls.java_processes.pop(port, None)
        if p is None:
            warnings.warn(
                f"Attempting to force-kill a genius brdige we did not start "
                "at port {port}",
                warnings.NegmasBridgeProcessWarning,
            )
            return False
        _kill_process(p.pid)
        if wait:
            cls.wait_until_not_listening(port)
        cls._close_gateway(port)
        return True

    @classmethod
    def stop(cls, port: int = DEFAULT_JAVA_PORT) -> bool:
        """Stops a running bridge

        Args:
            port: port number to use

        Returns:
            True if successful

        Remarks:
            - You should use this method to stop bridges.
            - It tries the following in order:
                1. shutdown the java bridge by calling its shutdown() method.
                2. killing the java bridge by calling its kill() method.
                3. killing the java bridge forcibly by killing the process
            - This method always waits for a short time to allow
              each process to complete. If it returns True then
              the bridge is no longer listening on the given port.

        """
        assert port > 0
        if not genius_bridge_is_running(port):
            return True
        cls.shutdown(port, wait=True)
        if not genius_bridge_is_running(port):
            return True
        cls.kill(port, wait=True)
        if not genius_bridge_is_running(port):
            return True
        cls.kill_forced(port, wait=True)
        if not genius_bridge_is_running(port):
            return True
        return False

    @classmethod
    def restart(cls, port: int = DEFAULT_JAVA_PORT, *args, **kwargs) -> bool:
        """Starts or restarts the genius bridge

        Args:
            port: port number to use

        Returns:
            True if successful

        """
        assert port > 0
        if not cls.stop(port):
            return False
        cls.wait_until_not_listening(port, 1)
        if not cls.start(port, *args, **kwargs):
            return False
        cls.wait_until_listening(port, 1)
        return genius_bridge_is_running(port)

    @classmethod
    def kill_threads(
        cls,
        port: int = DEFAULT_JAVA_PORT,
        wait_time: float = 0.5,
    ) -> bool:
        """kills all threads in the given java bridge"""
        assert port > 0
        try:
            gateway = cls.gateway(port, force=True)
        except Exception as e:
            warnings.warn(
                f"Failed to kill threads at port {port} with error: {str(e)}\n"
                f"{traceback.format_exc()}",
                warnings.NegmasBridgeProcessWarning,
            )
            return False
        if gateway is None:
            return False
        try:
            gateway.entry_point.kill_threads(int(wait_time * 1000))
        except:
            pass
        return True

    @classmethod
    def clean(cls, port=DEFAULT_JAVA_PORT) -> bool:
        """
        Removes all agents and runs garbage collection on the bridge
        """
        assert port > 0
        try:
            gateway = cls.gateway(port, force=True)
        except Exception as e:
            warnings.warn(
                f"Failed to kill threads at port {port} with error: {str(e)}\n"
                f"{traceback.format_exc()}",
                warnings.NegmasBridgeProcessWarning,
            )
            return False
        if gateway is None:
            return False
        gateway.entry_point.clean()
        return True

    @classmethod
    def clean_all(cls) -> bool:
        """
        Removes all agents and runs garbage collection on all bridges.
        """
        success = True
        for port in cls.gateways.keys():
            success = success and cls.clean(port)
        return success

    @classmethod
    def connect(cls, port: int = DEFAULT_JAVA_PORT) -> JavaObject:
        """
        Connects to a running genius-bridge

        Args:
            port: The port at which the bridge in listening in Java

        Remarks:
            - The difference between this method and start() is that this one
              does not attempt to start a java bridge if one does not exist.

        """
        assert port > 0
        try:
            gateway = cls.gateway(port, force=True)
        except Exception as e:
            warnings.warn(
                f"Failed to kill threads at port {port} with error: {str(e)}\n"
                f"{traceback.format_exc()}",
                warnings.NegmasBridgeProcessWarning,
            )
            raise RuntimeError(e)
        if gateway is None:
            raise RuntimeError(f"Got None as the gateway!!")
        return gateway.entry_point



from __future__ import annotations

import math
from typing import TYPE_CHECKING, Iterable, TypeVar

import numpy as np
from numpy.ma.core import sqrt

from negmas import warnings
from negmas.outcomes import Issue, Outcome, discretize_and_enumerate_issues
from negmas.outcomes.common import os_or_none
from negmas.outcomes.issue_ops import enumerate_issues
from negmas.outcomes.protocols import OutcomeSpace

from .base_ufun import BaseUtilityFunction

if TYPE_CHECKING:

    from negmas.preferences.prob_ufun import ProbUtilityFunction

    from .base_ufun import BaseUtilityFunction
    from .crisp_ufun import UtilityFunction
    from .discounted import DiscountedUtilityFunction

    UFunType = TypeVar("UFunType", UtilityFunction, ProbUtilityFunction)

__all__ = [
    "pareto_frontier",
    "nash_point",
    "make_discounted_ufun",
    "scale_max",
    "normalize",
    "sample_outcome_with_utility",
    "extreme_outcomes",
    "minmax",
    "conflict_level",
    "opposition_level",
    "winwin_level",
]


def make_discounted_ufun(
    ufun: UFunType,
    cost_per_round: float | None = None,
    power_per_round: float | None = None,
    discount_per_round: float | None = None,
    cost_per_relative_time: float | None = None,
    power_per_relative_time: float | None = None,
    discount_per_relative_time: float | None = None,
    cost_per_real_time: float | None = None,
    power_per_real_time: float | None = None,
    discount_per_real_time: float | None = None,
    dynamic_reservation: bool = True,
) -> DiscountedUtilityFunction | UFunType:
    from negmas.preferences.discounted import ExpDiscountedUFun, LinDiscountedUFun

    if cost_per_round is not None and cost_per_round > 0.0:
        ufun = LinDiscountedUFun(  # type: ignore (discounted ufuns return values of the same tyupe as their base ufun)
            ufun=ufun,
            cost=cost_per_round,
            factor="step",
            power=power_per_round,
            dynamic_reservation=dynamic_reservation,
            name=ufun.name,
            outcome_space=ufun.outcome_space,
        )
    if cost_per_relative_time is not None and cost_per_relative_time > 0.0:
        ufun = LinDiscountedUFun(  # type: ignore (discounted ufuns return values of the same tyupe as their base ufun)
            ufun=ufun,
            cost=cost_per_relative_time,
            factor="relative_time",
            power=power_per_relative_time,
            dynamic_reservation=dynamic_reservation,
            name=ufun.name,
            outcome_space=ufun.outcome_space,
        )
    if cost_per_real_time is not None and cost_per_real_time > 0.0:
        ufun = LinDiscountedUFun(  # type: ignore (discounted ufuns return values of the same tyupe as their base ufun)
            ufun=ufun,
            cost=cost_per_real_time,
            factor="real_time",
            power=power_per_real_time,
            dynamic_reservation=dynamic_reservation,
            name=ufun.name,
            outcome_space=ufun.outcome_space,
        )
    if discount_per_round is not None and discount_per_round > 0.0:
        ufun = ExpDiscountedUFun(  # type: ignore (discounted ufuns return values of the same tyupe as their base ufun)
            ufun=ufun,
            discount=discount_per_round,
            factor="step",
            dynamic_reservation=dynamic_reservation,
            name=ufun.name,
            outcome_space=ufun.outcome_space,
        )
    if discount_per_relative_time is not None and discount_per_relative_time > 0.0:
        ufun = ExpDiscountedUFun(  # type: ignore (discounted ufuns return values of the same tyupe as their base ufun)
            ufun=ufun,
            discount=discount_per_relative_time,
            factor="relative_time",
            dynamic_reservation=dynamic_reservation,
            name=ufun.name,
            outcome_space=ufun.outcome_space,
        )
    if discount_per_real_time is not None and discount_per_real_time > 0.0:
        ufun = ExpDiscountedUFun(  # type: ignore (discounted ufuns return values of the same tyupe as their base ufun)
            ufun=ufun,
            discount=discount_per_real_time,
            factor="real_time",
            dynamic_reservation=dynamic_reservation,
            name=ufun.name,
            outcome_space=ufun.outcome_space,
        )
    return ufun


def _pareto_frontier(
    points, eps=-1e-18, sort_by_welfare=False
) -> tuple[list[tuple[float]], list[int]]:
    """Finds the pareto-frontier of a set of points

    Args:
        points: list of points
        eps: A (usually negative) small number to treat as zero during calculations
        sort_by_welfare: If True, the results are sorted descindingly by total welfare

    Returns:

    """
    points = np.asarray(points)
    n = len(points)
    indices = np.array(range(n))
    for j in range(points.shape[1]):
        order = points[:, 0].argsort()[-1::-1]
        points = points[order]
        indices = indices[order]

    frontier = [(indices[0], points[0, :])]
    for p in range(1, n):
        current = points[p, :]
        for i, (_, f) in enumerate(frontier):
            current_better, current_worse = current > f, current < f
            if np.all(current == f):
                break
            if not np.any(current_better) and np.any(current_worse):
                # current is dominated, break
                break
            if np.any(current_better):
                if not np.any(current_worse):
                    # current dominates f, append it, remove f and scan for anything else dominated by current
                    for j, (_, g) in enumerate(frontier[i + 1 :]):
                        if np.all(current == g):
                            frontier = frontier[:i] + frontier[i + 1 :]
                            break
                        if np.any(current > g) and not np.any(current < g):
                            frontier = frontier[:j] + frontier[j + 1 :]
                    else:
                        frontier[i] = (indices[p], current)
                else:
                    # neither current nor f dominate each other, append current only if it is not
                    # dominated by anything in frontier
                    for j, (_, g) in enumerate(frontier[i + 1 :]):
                        if np.all(current == g) or (
                            np.any(g > current) and not np.any(current > g)
                        ):
                            break
                    else:
                        frontier.append((indices[p], current))
    if sort_by_welfare:
        welfare = [np.sum(_[1]) for _ in frontier]
        indx = sorted(range(len(welfare)), key=lambda x: welfare[x], reverse=True)
        frontier = [frontier[_] for _ in indx]
    return [tuple(_[1]) for _ in frontier], [_[0] for _ in frontier]


def nash_point(
    ufuns: Iterable[UtilityFunction],
    frontier: Iterable[tuple[float]],
    outcome_space: OutcomeSpace | None = None,
    issues: tuple[Issue] | None = None,
    outcomes: tuple[Outcome] | None = None,
) -> tuple[tuple[float, ...] | None, int | None]:
    """
    Calculates the nash point on the pareto frontier of a negotiation

    Args:
        ufuns: A list of ufuns to use
        frontier: a list of tuples each giving the utility values at some outcome on the frontier (usually found by `pareto_frontier`) to search within
        outcome_space: The outcome-space to consider
        issues: The issues on which the ufun is defined (outcomes may be passed instead)
        outcomes: The outcomes on which the ufun is defined (outcomes may be passed instead)

    Returns:

        A tuple of three values (all will be None if reserved values are unknown)

        - A tuple of utility values at the nash point
        - The index of the given frontier corresponding to the nash point

    Remarks:

        - The function searches within the given frontier only.

    """
    nash_val = float("-inf")
    u_vals = None
    nash_indx = None
    ranges = [_.minmax(outcome_space, issues, outcomes) for _ in ufuns]
    for i, (rng, ufun) in enumerate(zip(ranges, ufuns)):
        if any(_ is None or not math.isfinite(_) for _ in rng):
            return None, None
        r = ufun.reserved_value
        if r is None or r < rng[0]:
            continue
        ranges[i] = (r, rng[1])
    if any([_[1] <= 1.0e-9 for _ in ranges]):
        return None, None
    diffs = [float(b) - float(r) for r, b in ranges]
    for indx, us in enumerate(frontier):
        val = 0.0
        for u, (r, _), d in zip(us, ranges, diffs):
            val *= (float(u) - float(r)) / d
        if val > nash_val:
            nash_val = val
            u_vals = us
            nash_indx = indx
    return u_vals, nash_indx


def pareto_frontier(
    ufuns: Iterable[UtilityFunction],
    outcomes: Iterable[Outcome] | None = None,
    issues: Iterable[Issue] | None = None,
    n_discretization: int | None = 10,
    sort_by_welfare=False,
) -> tuple[list[tuple[float, ...]], list[int]]:
    """Finds all pareto-optimal outcomes in the list

    Args:

        ufuns: The utility functions
        outcomes: the outcomes to be checked. If None then all possible outcomes from the issues will be checked
        issues: The set of issues (only used when outcomes is None)
        n_discretization: The number of items to discretize each real-dimension into
        sort_by_welfare: If True, the resutls are sorted descendingly by total welfare

    Returns:
        Two lists of the same length. First list gives the utilities at pareto frontier points and second list gives their indices

    """

    ufuns = tuple(ufuns)
    if issues:
        issues = tuple(issues)
    if outcomes:
        outcomes = tuple(outcomes)

    # calculate all candidate outcomes
    if outcomes is None:
        if issues is None:
            return [], []
        outcomes = discretize_and_enumerate_issues(issues, n_discretization)
        # outcomes = itertools.product(
        #     *[issue.value_generator(n=n_discretization) for issue in issues]
        # )
    points = [[ufun(outcome) for ufun in ufuns] for outcome in outcomes]
    return _pareto_frontier(points, sort_by_welfare=sort_by_welfare)


def scale_max(
    ufun: UFunType,
    to: float = 1.0,
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
) -> UFunType:
    """Normalizes a utility function to the given range

    Args:
        ufun: The utility function to normalize
        outcomes: A collection of outcomes to normalize for
        rng: range to normalize to. Default is [0, 1]
        epsilon: A small number specifying the resolution

    Returns:
        UtilityFunction: A utility function that is guaranteed to be normalized for the set of given outcomes

    """
    return ufun.scale_max_for(
        to, issues=issues, outcome_space=outcome_space, outcomes=outcomes
    )


def normalize(
    ufun: BaseUtilityFunction,
    to: tuple[float, float] = (0.0, 1.0),
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
) -> BaseUtilityFunction:
    """Normalizes a utility function to the given range

    Args:
        ufun: The utility function to normalize
        to: range to normalize to. Default is [0, 1]
        outcomes: A collection of outcomes to normalize for

    Returns:
        UtilityFunction: A utility function that is guaranteed to be normalized for the set of given outcomes

    """
    outcome_space = os_or_none(outcome_space, issues, outcomes)
    if outcome_space is None:
        return ufun.normalize(to)
    return ufun.normalize_for(to, outcome_space=outcome_space)


def sample_outcome_with_utility(
    ufun: BaseUtilityFunction,
    rng: tuple[float, float],
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
    n_trials: int = 100,
) -> Outcome | None:
    """
    Gets one outcome within the given utility range or None on failure

    Args:
        ufun: The utility function
        rng: The utility range
        outcome_space: The outcome-space within which to sample
        issues: The issues the utility function is defined on
        outcomes: The outcomes to sample from
        n_trials: The maximum number of trials

    Returns:

        - Either issues, or outcomes should be given but not both

    """
    return ufun.sample_outcome_with_utility(
        rng, outcome_space, issues, outcomes, n_trials
    )


def extreme_outcomes(
    ufun: UtilityFunction,
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
    max_cardinality=1000,
) -> tuple[Outcome, Outcome]:
    """
    Finds the best and worst outcomes

    Args:
        ufun: The utility function
        outcome_space: An outcome-space to consider
        issues: list of issues (optional)
        outcomes: A collection of outcomes (optional)
        max_cardinality: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not
                        given)
    Returns:
        Outcomes with minumum utility, maximum utility

    """
    return ufun.extreme_outcomes(
        outcome_space=outcome_space,
        issues=issues,
        outcomes=outcomes,
        max_cardinality=max_cardinality,
    )


def minmax(
    ufun: UtilityFunction,
    outcome_space: OutcomeSpace | None = None,
    issues: list[Issue] | None = None,
    outcomes: list[Outcome] | None = None,
    max_cardinality=1000,
) -> tuple[float, float]:
    """Finds the range of the given utility function for the given outcomes

    Args:
        ufun: The utility function
        outcome_space: An outcome-space to consider
        issues: list of issues (optional)
        outcomes: A collection of outcomes (optional)
        max_cardinality: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not
                        given)
    Returns:
        Minumum utility, maximum utility

    """
    return ufun.minmax(
        outcome_space=outcome_space,
        issues=issues,
        outcomes=outcomes,
        max_cardinality=max_cardinality,
    )


def opposition_level(
    ufuns: list[UtilityFunction] | tuple[UtilityFunction, ...],
    max_utils: float | tuple[float, float] = 1.0,  # type: ignore
    outcomes: int | list[Outcome] | tuple[Outcome, ...] | None = None,
    issues: list[Issue] | None = None,
    max_tests: int = 10000,
) -> float:
    """
    Finds the opposition level of the two ufuns defined as the minimum distance to outcome (1, 1)

    Args:
        ufuns: A list of utility functions to use.
        max_utils: A list of maximum utility value for each ufun (or a single number if they are equal).
        outcomes: A list of outcomes (should be the complete issue space) or an integer giving the number
                 of outcomes. In the later case, ufuns should expect a tuple of a single integer.
        issues: The issues (only used if outcomes is None).
        max_tests: The maximum number of outcomes to use. Only used if issues is given and has more
                   outcomes than this value.


    Examples:


        - Opposition level of the same ufun repeated is always 0
        >>> from negmas.preferences.crisp.mapping import MappingUtilityFunction
        >>> from negmas.preferences.ops import opposition_level
        >>> u1, u2 = lambda x: x[0], lambda x: x[0]
        >>> opposition_level([u1, u2], outcomes=10, max_utils=9)
        0.0

        - Opposition level of two ufuns that are zero-sum
        >>> u1, u2 = MappingUtilityFunction(lambda x: x[0]), MappingUtilityFunction(lambda x: 9 - x[0])
        >>> opposition_level([u1, u2], outcomes=10, max_utils=9)
        0.7114582486036499

    """
    if outcomes is None:
        if issues is None:
            raise ValueError("You must either give outcomes or issues")
        outcomes = list(enumerate_issues(tuple(issues), max_cardinality=max_tests))
    if isinstance(outcomes, int):
        outcomes = [(_,) for _ in range(outcomes)]
    if not isinstance(max_utils, Iterable):
        max_utils: Iterable[float] = [max_utils] * len(ufuns)
    if len(ufuns) != len(max_utils):
        raise ValueError(
            f"Cannot use {len(ufuns)} ufuns with only {len(max_utils)} max. utility values"
        )

    nearest_val = float("inf")
    for outcome in outcomes:
        v = sum(
            (1.0 - float(u(outcome)) / max_util) ** 2
            if max_util
            else (1.0 - float(u(outcome))) ** 2
            for max_util, u in zip(max_utils, ufuns)
        )
        if v == float("inf"):
            warnings.warn(
                f"u is infinity: {outcome}, {[_(outcome) for _ in ufuns]}, max_utils",
                warnings.NegmasNumericWarning,
            )
        if v < nearest_val:
            nearest_val = v
    return sqrt(nearest_val)


def conflict_level(
    u1: UtilityFunction,
    u2: UtilityFunction,
    outcomes: int | list[Outcome] | tuple[Outcome],
    max_tests: int = 10000,
) -> float:
    """
    Finds the conflict level in these two ufuns

    Args:
        u1: first utility function
        u2: second utility function

    Examples:
        - A nonlinear strictly zero sum case
        >>> from negmas.preferences.crisp.mapping import MappingUtilityFunction
        >>> from negmas.preferences import conflict_level
        >>> outcomes = [(_,) for _ in range(10)]
        >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
        ... np.random.random(len(outcomes)))))
        >>> u2 = MappingUtilityFunction(dict(zip(outcomes,
        ... 1.0 - np.array(list(u1.mapping.values())))))
        >>> print(conflict_level(u1=u1, u2=u2, outcomes=outcomes))
        1.0

        - The same ufun
        >>> print(conflict_level(u1=u1, u2=u1, outcomes=outcomes))
        0.0

        - A linear strictly zero sum case
        >>> outcomes = [(i,) for i in range(10)]
        >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
        ... np.linspace(0.0, 1.0, len(outcomes), endpoint=True))))
        >>> u2 = MappingUtilityFunction(dict(zip(outcomes,
        ... np.linspace(1.0, 0.0, len(outcomes), endpoint=True))))
        >>> print(conflict_level(u1=u1, u2=u2, outcomes=outcomes))
        1.0
    """
    if isinstance(outcomes, int):
        outcomes = [(_,) for _ in range(outcomes)]
    else:
        outcomes = list(outcomes)
    n_outcomes = len(outcomes)
    if n_outcomes == 0:
        raise ValueError(f"Cannot calculate conflit level with no outcomes")
    points = np.array([[u1(o), u2(o)] for o in outcomes])
    order = np.random.permutation(np.array(range(n_outcomes)))
    p1, p2 = points[order, 0], points[order, 1]
    signs = []
    trial = 0
    for i in range(n_outcomes - 1):
        for j in range(i + 1, n_outcomes):
            if trial >= max_tests:
                break
            trial += 1
            o11, o12 = p1[i], p1[j]
            o21, o22 = p2[i], p2[j]
            if o12 == o11 and o21 == o22:
                continue
            signs.append(int((o12 > o11 and o21 > o22) or (o12 < o11 and o21 < o22)))
    signs = np.array(signs)
    # todo: confirm this is correct
    if len(signs) == 0:
        return 1.0
    return signs.mean()


def winwin_level(
    u1: UtilityFunction,
    u2: UtilityFunction,
    outcomes: int | list[Outcome] | tuple[Outcome],
    max_tests: int = 10000,
) -> float:
    """
    Finds the win-win level in these two ufuns

    Args:
        u1: first utility function
        u2: second utility function

    Examples:
        - A nonlinear same ufun case
        >>> from negmas.preferences.crisp.mapping import MappingUtilityFunction
        >>> outcomes = [(_,) for _ in range(10)]
        >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
        ... np.linspace(1.0, 0.0, len(outcomes), endpoint=True))))

        - A linear strictly zero sum case
        >>> outcomes = [(_,) for _ in range(10)]
        >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
        ... np.linspace(0.0, 1.0, len(outcomes), endpoint=True))))
        >>> u2 = MappingUtilityFunction(dict(zip(outcomes,
        ... np.linspace(1.0, 0.0, len(outcomes), endpoint=True))))


    """
    if isinstance(outcomes, int):
        outcomes = [(_,) for _ in range(outcomes)]
    else:
        outcomes = list(outcomes)
    n_outcomes = len(outcomes)
    points = np.array([[u1(o), u2(o)] for o in outcomes])
    order = np.random.permutation(np.array(range(n_outcomes)))
    p1, p2 = points[order, 0], points[order, 1]
    signed_diffs = []
    for trial, (i, j) in enumerate(zip(range(n_outcomes - 1), range(1, n_outcomes))):
        if trial >= max_tests:
            break
        o11, o12 = p1[i], p1[j]
        o21, o22 = p2[i], p2[j]
        if o11 == o12:
            if o21 == o22:
                continue
            else:
                win = abs(o22 - o21)
        elif o11 < o12:
            if o21 == o22:
                win = o12 - o11
            else:
                win = (o12 - o11) + (o22 - o21)
        else:
            if o21 == o22:
                win = o11 - o12
            else:
                win = (o11 - o12) + (o22 - o21)
        signed_diffs.append(win)
    signed_diffs = np.array(signed_diffs)
    if len(signed_diffs) == 0:
        raise ValueError("Could not calculate any signs")
    return signed_diffs.mean()



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from negmas import Value
from negmas.preferences.base_ufun import BaseUtilityFunction
from negmas.preferences.preferences import Preferences
from negmas.warnings import NegmasUnexpectedValueWarning, warn

from ...events import Notification
from ...negotiators import Controller, Negotiator
from ...outcomes import Outcome
from ..common import ResponseType, SAOResponse

if TYPE_CHECKING:
    from negmas.sao import SAOState
    from negmas.situated import Agent

__all__ = [
    "SAONegotiator",
]


class SAONegotiator(Negotiator):
    """
    Base class for all SAO negotiators.

    Args:
         name: Negotiator name
         parent: Parent controller if any
         preferences: The preferences of the negotiator
         ufun: The utility function of the negotiator (overrides preferences if given)
         owner: The `Agent` that owns the negotiator.

    Remarks:
        - The only method that **must** be implemented by any SAONegotiator is `propose`.
        - The default `respond` method, accepts offers with a utility value no less than whatever `propose` returns
          with the same mechanism state.

    """

    def __init__(
        self,
        preferences: Preferences | None = None,
        ufun: BaseUtilityFunction | None = None,
        name: str | None = None,
        parent: Controller | None = None,
        owner: Agent | None = None,
        id: str | None = None,
        type_name: str | None = None,
        can_propose: bool = True,
    ):
        super().__init__(
            name=name,
            preferences=preferences,
            ufun=ufun,
            parent=parent,
            owner=owner,
            id=id,
            type_name=type_name,
        )
        self.__end_negotiation = False
        self.my_last_proposal: Outcome | None = None
        self._my_last_proposal_utility: Value | None = None
        self.add_capabilities(
            {"respond": True, "propose": can_propose, "max-proposals": 1}
        )

    def on_notification(self, notification: Notification, notifier: str):
        """
        Called whenever a notification is received

        Args:
            notification: The notification
            notifier: The notifier entity

        Remarks:
            - The default implementation only responds to end_negotiation by ending the negotiation
        """
        if notification.type == "end_negotiation":
            self.__end_negotiation = True

    @abstractmethod
    def propose(self, state: SAOState) -> Outcome | None:
        """Propose an offer or None to refuse.

        Args:
            state: `SAOState` giving current state of the negotiation.

        Returns:
            The outcome being proposed or None to refuse to propose

        Remarks:
            - This function guarantees that no agents can propose something with a utility value

        """

    def propose_(self, state: SAOState) -> Outcome | None:
        """
        The method directly called by the mechanism (through `counter` ) to ask for a proposal

        Args:
            state: The mechanism state

        Returns:
            An outcome to offer or None to refuse to offer

        Remarks:
            - Depending on the `SAOMechanism` settings, refusing to offer may be interpreted as ending the negotiation
            - The negotiator will only receive this call if it has the 'propose' capability.

        """
        if not self._capabilities["propose"] or self.__end_negotiation:
            return None
        if not state.running:
            warn(
                f"{self.name} asked to propose in a negotiation that is not running:\n{state}"
            )
            return None
        return self.propose(state=state)

    def respond(self, state: SAOState, offer: Outcome) -> ResponseType:
        """Called to respond to an offer. This is the method that should be overriden to provide an acceptance strategy.

        Args:
            state: a `SAOState` giving current state of the negotiation.
            offer: offer being tested

        Returns:
            ResponseType: The response to the offer

        Remarks:
            - The default implementation never ends the negotiation
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        # Always reject offers that we do not know or if we have no kknown preferences
        if offer is None or self.ufun is None:
            return ResponseType.REJECT_OFFER
        # find offer utility
        offer_util = self.ufun(offer) if self.ufun else float("-inf")
        # if the offer is worse than the reserved value or its utility is less than that of the reserved outcome, reject it
        if (self.reserved_value is not None and offer_util < self.reserved_value) or (
            self.reserved_outcome is not None
            and self.ufun is not None
            and offer_util < self.ufun(self.reserved_outcome)
        ):
            return ResponseType.REJECT_OFFER
        # find my last proposal (if any)
        utility = None
        if self._my_last_proposal_utility is not None:
            utility = self._my_last_proposal_utility
        # if I never proposed, find what would I have proposed at this state and its utility
        if utility is None:
            myoffer = self.propose_(state=state)
            if myoffer is None:
                return ResponseType.NO_RESPONSE
            utility = self.ufun(myoffer)
        # accept only if I know what I would have proposed at this state (or the previous one) and it was worse than what I am about to proposed
        if utility is not None and offer_util >= utility:
            return ResponseType.ACCEPT_OFFER
        return ResponseType.REJECT_OFFER

    def respond_(self, state: SAOState, offer: Outcome) -> ResponseType:
        """The method to be called directly by the mechanism (through `counter` ) to respond to an offer.

        Args:
            state: a `SAOState` giving current state of the negotiation.
            offer: the offer being responded to.

        Returns:
            ResponseType: The response to the offer. Possible values are:

                - NO_RESPONSE: refuse to offer. Depending on the mechanism settings this may be interpreted as ending
                               the negotiation.
                - ACCEPT_OFFER: Accepting the offer.
                - REJECT_OFFER: Rejecting the offer. The negotiator will be given the chance to counter this
                                offer through a call of `propose_` later if this was not the last offer to be evaluated
                                by the mechanism.
                - END_NEGOTIATION: End the negotiation
                - WAIT: Instructs the mechanism to wait for this negotiator more. It may lead to cycles so use with care.

        Remarks:
            - The default implementation never ends the negotiation except if an earler end_negotiation notification is
              sent to the negotiator
            - The default implementation asks the negotiator to `propose`() and accepts the `offer` if its utility was
              at least as good as the offer that it would have proposed (and above the reserved value).

        """
        if not state.running:
            warn(
                f"{self.name} asked to respond to a negotiation that is not running:\n{state}"
            )
            return ResponseType.END_NEGOTIATION
        if self.__end_negotiation:
            return ResponseType.END_NEGOTIATION
        return self.respond(state=state, offer=offer)

    def counter(self, state: SAOState, offer: Outcome | None) -> SAOResponse:
        """
        Called by the mechanism to counter the offer. It just calls `respond_` and `propose_` as needed.

        Args:
            state: `SAOState` giving current state of the negotiation.
            offer: The offer to be countered. None means no offer and the agent is requested to propose an offer

        Returns:
            Tuple[ResponseType, Outcome]: The response to the given offer with a counter offer if the response is REJECT

        """
        if self.__end_negotiation:
            return SAOResponse(ResponseType.END_NEGOTIATION, None)
        if self.ufun is not None:
            changes = self.ufun.changes()
            if changes:
                self.on_preferences_changed(changes)
        if offer is None:
            return SAOResponse(ResponseType.REJECT_OFFER, self.propose_(state=state))
        response = self.respond_(state=state, offer=offer)
        if response != ResponseType.REJECT_OFFER:
            return SAOResponse(response, None)
        return SAOResponse(response, self.propose_(state=state))

    def on_partner_proposal(
        self, state: SAOState, partner_id: str, offer: Outcome
    ) -> None:
        """
        A callback called by the mechanism when a partner proposes something

        Args:
            state: `SAOState` giving the state of the negotiation when the offer was porposed.
            partner_id: The ID of the agent who proposed
            offer: The proposal.

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_refused_to_propose(self, state: SAOState, partner_id: str) -> None:
        """
        A callback called by the mechanism when a partner refuses to propose

        Args:
            state: `SAOState` giving the state of the negotiation when the partner refused to offer.
            partner_id: The ID of the agent who refused to propose

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """

    def on_partner_response(
        self,
        state: SAOState,
        partner_id: str,
        outcome: Outcome,
        response: ResponseType,
    ) -> None:
        """
        A callback called by the mechanism when a partner responds to some offer

        Args:
            state: `SAOState` giving the state of the negotiation when the partner responded.
            partner_id: The ID of the agent who responded
            outcome: The proposal being responded to.
            response: The response

        Remarks:
            - Will only be called if `enable_callbacks` is set for the mechanism
        """



from __future__ import annotations

import math
import numbers
import random
from typing import Generator

import numpy as np

from negmas.outcomes.base_issue import Issue
from negmas.outcomes.range_issue import RangeIssue

from .common import DEFAULT_LEVELS

__all__ = ["ContinuousIssue"]


class ContinuousIssue(RangeIssue):
    """
    A `RangeIssue` representing a continous range of real numbers with finite limits.
    """

    def __init__(self, values, name=None, n_levels=DEFAULT_LEVELS) -> None:
        super().__init__(values, name=name)
        self._n_levels = n_levels
        self.delta = (self.max_value - self.min_value) / self._n_levels

    def _to_xml_str(self, indx):
        return (
            f'    <issue etype="real" index="{indx + 1}" name="{self.name}" type="real" vtype="real">\n'
            f'        <range lowerbound="{self._values[0]}" upperbound="{self._values[1]}"></range>\n    </issue>\n'
        )

    def to_dict(self):
        d = super().to_dict()
        d["n_levels"] = self._n_levels
        return d

    @property
    def cardinality(self) -> int | float:
        return float("inf")

    @property
    def type(self) -> str:
        return "continuous"

    def is_continuous(self) -> bool:
        return True

    def is_uncountable(self) -> bool:
        return True

    def rand(self) -> float:
        """Picks a random valid value."""

        return (
            random.random() * (self._values[1] - self._values[0]) + self._values[0]
        )  # type: ignore

    def ordered_value_generator(
        self,
        n: int | float | None = DEFAULT_LEVELS,
        grid=True,
        compact=False,
        endpoints=True,
    ) -> Generator[float, None, None]:
        if n is None or not math.isfinite(n):
            raise ValueError(f"Cannot generate {n} values from issue: {self}")
        n = int(n)
        if grid:
            yield from np.linspace(
                self._values[0], self._values[1], num=n, endpoint=endpoints
            ).tolist()
            return
        if endpoints:
            yield from [self._values[0]] + (
                (self._values[1] - self._values[0]) * np.random.rand(n - 2)
                + self._values[0]
            ).tolist() + [self._values[1]]
        yield from (
            (self._values[1] - self._values[0]) * np.random.rand(n) + self._values[0]
        ).tolist()

    def value_generator(
        self,
        n: int | float | None = DEFAULT_LEVELS,
        grid=True,
        compact=False,
        endpoints=True,
    ) -> Generator[float, None, None]:
        return self.ordered_value_generator(n, grid, compact, endpoints)

    def rand_outcomes(
        self, n: int, with_replacement=False, fail_if_not_enough=False
    ) -> list[float]:
        if with_replacement:
            return (
                np.random.rand(n) * (self._values[1] - self._values[0])
                + self._values[0]
            ).tolist()
        return np.linspace(
            self._values[0], self._values[1], num=n, endpoint=True
        ).tolist()

    def rand_invalid(self):
        """Pick a random *invalid* value"""

        return (
            random.random() * (self.max_value - self.min_value) + self.max_value * 1.1
        )

    @property
    def all(self):
        raise ValueError(f"Cannot enumerate all values of a continuous issue")

    def value_at(self, index: int):
        v = self.min_value + self.delta * index
        if v > self.max_value:
            return IndexError(index)
        return v

    def contains(self, issue: Issue) -> bool:
        """Checks weather this issue contains the input issue (i.e. every value in the input issue is in this issue)"""
        return (
            issubclass(issue.value_type, numbers.Real)
            and issue.min_value >= self.min_value
            and issue.max_value <= self.max_value
        )



from __future__ import annotations

import pkg_resources
import pytest

from negmas.genius.ginfo import ALL_PASSING_NEGOTIATORS as ALL_NEGOTIATORS
from negmas.genius.gnegotiators import (
    ABMPAgent2,
    AgentBuyong,
    AgentHP2,
    AgentK,
    AgentK2,
    AgentLG,
    AgentM,
    AgentX,
    AgentYK,
    AgreeableAgent2018,
    Atlas3,
    Atlas32016,
    BayesianAgent,
    BetaOne,
    BoulwareNegotiationParty,
    BRAMAgent,
    Caduceus,
    CaduceusDC16,
    ConcederNegotiationParty,
    CUHKAgent,
    DoNA,
    E2Agent,
    Farma,
    FuzzyAgent,
    Gahboninho,
    Gangster,
    GeniusNegotiator,
    GrandmaAgent,
    Group2,
    HardHeaded,
    IAMhaggler,
    IAMhaggler2011,
    ImmediateAcceptor,
    Kawaii,
    KGAgent,
    MengWan,
    MetaAgent,
    MyAgent,
    Ngent,
    NiceTitForTat,
    Nozomi,
    OMACagent,
    OptimalBidderSimple,
    ParsAgent,
    ParsCat,
    PhoenixParty,
    PokerFace,
    PonPokoAgent,
    RandomCounterOfferNegotiationParty,
    RandomDance,
    RandomParty,
    RandomParty2,
    Rubick,
    SimilarityAgent,
    Simpatico,
    SimpleAgent,
    Terra,
    TheFawkes,
    TheNegotiator,
    TheNegotiatorReloaded,
    TimeDependentAgentBoulware,
    TimeDependentAgentConceder,
    TimeDependentAgentHardliner,
    TimeDependentAgentLinear,
    TMFAgent,
    ValueModelAgent,
    WhaleAgent,
    XianFaAgent,
    Yushu,
    YXAgent,
)
from negmas.inout import Scenario
from negmas.sao.negotiators import ToughNegotiator

from .switches import NEGMAS_FASTRUN, NEGMAS_RUN_GENIUS

TIMELIMIT = 30
STEPLIMIT = 50

AGENTS_WITH_NO_AGREEMENT_ON_SAME_preferences = (
    "agents.anac.y2015.Mercury.Mercury",
    "parties.in4010.q12015.group19.Group19",
    "parties.in4010.q12015.group6.Group6",
    "agents.anac.y2015.xianfa.XianFaAgent",
)

DOMAINS = [
    "tests/data/Car-A-domain",
    # "tests/data/Laptop",
]

SKIP_CONDITION = NEGMAS_FASTRUN or not NEGMAS_RUN_GENIUS


def do_run(
    agent_factory,
    base_folder,
    opponent_preferences,
    agent_preferences,
    agent_starts,
    opponent_factory,
    n_steps,
    time_limit,
):
    domain = Scenario.from_genius_folder(base_folder)
    if not domain:
        raise ValueError(f"Cannot open domain {base_folder}")
    neg = domain.make_session(
        n_steps=n_steps, time_limit=time_limit, avoid_ultimatum=False
    )
    if neg is None:
        raise ValueError(f"Failed to load domain from {base_folder}")
    opponent = opponent_factory(preferences=domain.ufuns[opponent_preferences])
    theagent = agent_factory(preferences=domain.ufuns[agent_preferences])
    if agent_starts:
        neg.add(theagent)
        neg.add(opponent)
    else:
        neg.add(opponent)
        neg.add(theagent)
    neg.run()
    return neg


def do_test_same_ufun(agent_factory, base_folder, n_steps, time_limit, n_trials=3):
    # check that it will get to an agreement sometimes if the same ufun
    # is used for both agents
    from random import randint

    for _ in range(n_trials):
        indx = randint(0, 1)
        neg = do_run(
            agent_factory,
            base_folder,
            indx,
            indx,
            False,
            agent_factory,
            n_steps,
            time_limit,
        )
        if neg.agreement is not None:
            break
    else:
        assert (
            False
        ), f"failed to get an agreement in {n_trials} trials even using the same ufun\n{neg.trace}"  # type: ignore It makes not sense to have n_trials == 0 so we are safe


def do_test_genius_agent(
    AgentFactory, must_agree_if_same_preferences=True, java_class_name=None
):
    if java_class_name is not None:
        AgentFactory = lambda *args, **kwargs: GeniusNegotiator(
            *args, java_class_name=java_class_name, **kwargs
        )
        agent_class_name = java_class_name
    else:
        agent_class_name = AgentFactory.__name__
    # print(f"Running {AgentClass.__name__}")
    for domain in DOMAINS:
        base_folder = pkg_resources.resource_filename("negmas", resource_name=domain)

        # check that it can run without errors with two different ufuns
        for opponent_type in (ToughNegotiator, Atlas3):
            for starts in (False, True):
                for n_steps, time_limit in ((STEPLIMIT, None), (None, TIMELIMIT)):
                    for ufuns in ((1, 0), (0, 1)):
                        try:
                            do_run(
                                AgentFactory,
                                base_folder,
                                ufuns[0],
                                ufuns[1],
                                starts,
                                opponent_type,
                                n_steps=n_steps,
                                time_limit=time_limit,
                            )
                        except Exception as e:
                            print(
                                f"{agent_class_name} FAILED against {opponent_type.__name__}"
                                f" going {'first' if starts else 'last'} ({n_steps} steps with "
                                f"{time_limit} limit taking ufun {ufuns[1]}."
                            )
                            raise e

        if not must_agree_if_same_preferences or (
            java_class_name is None
            and AgentFactory in AGENTS_WITH_NO_AGREEMENT_ON_SAME_preferences
        ):
            continue
        do_test_same_ufun(AgentFactory, base_folder, STEPLIMIT, None, 3)

    # GeniusBridge.clean()


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
@pytest.mark.parametrize("negotiator", ALL_NEGOTIATORS)
def test_all_negotiators(negotiator):
    do_test_genius_agent(None, java_class_name=negotiator)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_boulware_party():
    do_test_genius_agent(
        None, java_class_name="negotiator.parties.BoulwareNegotiationParty"
    )


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_boulware_agent():
    do_test_genius_agent(None, java_class_name="agents.TimeDependentAgentBoulware")


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentX():
    do_test_genius_agent(AgentX)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_YXAgent():
    do_test_genius_agent(YXAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Caduceus():
    do_test_genius_agent(Caduceus)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_ParsCat():
    do_test_genius_agent(ParsCat)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_ParsAgent():
    do_test_genius_agent(ParsAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_PonPokoAgent():
    do_test_genius_agent(PonPokoAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_RandomDance():
    do_test_genius_agent(RandomDance)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Atlas32016():
    do_test_genius_agent(Atlas32016)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_MyAgent():
    do_test_genius_agent(MyAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Farma():
    do_test_genius_agent(Farma)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_PokerFace():
    do_test_genius_agent(PokerFace)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentBuyong():
    do_test_genius_agent(AgentBuyong)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Kawaii():
    do_test_genius_agent(Kawaii)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Atlas3():
    do_test_genius_agent(Atlas3)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentYK():
    do_test_genius_agent(AgentYK)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Group2():
    do_test_genius_agent(Group2)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_WhaleAgent():
    do_test_genius_agent(WhaleAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_DoNA():
    do_test_genius_agent(DoNA)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentM():
    do_test_genius_agent(AgentM)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_TMFAgent():
    do_test_genius_agent(TMFAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_OMACagent():
    do_test_genius_agent(OMACagent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentLG():
    do_test_genius_agent(AgentLG)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_CUHKAgent():
    do_test_genius_agent(CUHKAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_ValueModelAgent():
    do_test_genius_agent(ValueModelAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_NiceTitForTat():
    do_test_genius_agent(NiceTitForTat)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_TheNegotiator():
    do_test_genius_agent(TheNegotiator)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentK2():
    do_test_genius_agent(AgentK2)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_BRAMAgent():
    do_test_genius_agent(BRAMAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_IAMhaggler2011():
    do_test_genius_agent(IAMhaggler2011)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Gahboninho():
    do_test_genius_agent(Gahboninho)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_HardHeaded():
    do_test_genius_agent(HardHeaded)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_AgentK():
    do_test_genius_agent(AgentK)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Yushu():
    do_test_genius_agent(Yushu)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Nozomi():
    do_test_genius_agent(Nozomi)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_IAMhaggler():
    do_test_genius_agent(IAMhaggler)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Terra():
    do_test_genius_agent(Terra)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Gangster():
    do_test_genius_agent(Gangster)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_TheFawkes():
    do_test_genius_agent(TheFawkes)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_MetaAgent():
    do_test_genius_agent(MetaAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_TheNegotiatorReloaded():
    do_test_genius_agent(TheNegotiatorReloaded)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="Either no-genius or fast-run",
)
def test_Simpatico():
    do_test_genius_agent(Simpatico)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_ConcederNegotiationParty():
    do_test_genius_agent(ConcederNegotiationParty)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_BayesianAgent():
    do_test_genius_agent(BayesianAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_FuzzyAgent():
    do_test_genius_agent(FuzzyAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_ABMPAgent2():
    do_test_genius_agent(ABMPAgent2)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_OptimalBidderSimple():
    do_test_genius_agent(OptimalBidderSimple)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_SimilarityAgent():
    do_test_genius_agent(SimilarityAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_BoulwareNegotiationParty():
    do_test_genius_agent(BoulwareNegotiationParty)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_TimeDependentAgentLinear():
    do_test_genius_agent(TimeDependentAgentLinear)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_TimeDependentAgentHardliner():
    do_test_genius_agent(TimeDependentAgentHardliner)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_TimeDependentAgentConceder():
    do_test_genius_agent(TimeDependentAgentConceder)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_TimeDependentAgentBoulware():
    do_test_genius_agent(TimeDependentAgentBoulware)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_RandomParty2():
    do_test_genius_agent(RandomParty2)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_RandomParty():
    do_test_genius_agent(RandomParty)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_RandomCounterOfferNegotiationParty():
    do_test_genius_agent(RandomCounterOfferNegotiationParty)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_SimpleAgent():
    do_test_genius_agent(SimpleAgent)


@pytest.mark.skipif(
    condition=SKIP_CONDITION,
    reason="No Genius Gridge, skipping genius-agent tests",
)
def test_ImmediateAcceptor():
    do_test_genius_agent(ImmediateAcceptor)


#### agents after this line are not very robust


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_Rubick():
    do_test_genius_agent(Rubick)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_CaduceusDC16():
    do_test_genius_agent(CaduceusDC16)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_BetaOne():
    do_test_genius_agent(BetaOne)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_AgreeableAgent2018():
    do_test_genius_agent(AgreeableAgent2018)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_MengWan():
    do_test_genius_agent(MengWan)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_GrandmaAgent():
    do_test_genius_agent(GrandmaAgent)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_XianFaAgent():
    do_test_genius_agent(XianFaAgent)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_PhoenixParty():
    do_test_genius_agent(PhoenixParty)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_AgentHP2():
    do_test_genius_agent(AgentHP2)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_KGAgent():
    do_test_genius_agent(KGAgent)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_E2Agent():
    do_test_genius_agent(E2Agent)


@pytest.mark.xfail(
    run=False,
    reason="Genius agent known not to work",
)
def test_Ngent():
    do_test_genius_agent(Ngent)


if __name__ == "__main__":
    pytest.main(args=[__file__])



#!/usr/bin/env python
"""
A set of utilities that can be used by agents developed for the platform.

This set of utlities can be extended but must be backward compatible for at
least two versions
"""
from __future__ import annotations

import random
from typing import Iterable, TypeVar

import numpy as np
from scipy.stats import tmean

T = TypeVar("T")

# INTTYPES = (int, np.int32, np.int64, np.int8, np.int16)
# REALTYPES = (
#     float,
#     np.float16,
#     np.float32,
#     np.float64,
# )

__all__ = [
    "get_one_float",
    "get_one_int",
    "make_range",
    "truncated_mean",
    "sample",
    "is_int_type",
    "is_float_type",
    "isint",
    "isreal",
]


def is_int_type(x):
    """
    Checks if the given type is an integer (cannot be a floating point type)
    """
    return np.issubdtype(x, np.integer)


def is_float_type(x):
    """
    Checks if the given type is a floating point type (it cannot be an integer type)
    """
    return np.issubdtype(x, np.floating)


def isint(x):
    """
    Is the given number is of an integer type
    """
    return isinstance(x, np.integer)


def isreal(x):
    """
    Is the given number is of a floating point type
    """
    return isinstance(x, np.floating)


def get_one_int(i: int | tuple[int, int]):
    if isinstance(i, int):
        return i
    return random.randint(*i)


def get_one_float(rng: float | tuple[float, float]):
    if isinstance(rng, float):
        return rng
    return random.random() * (rng[1] - rng[0]) + rng[0]


def make_range(x: T | tuple[T, T]) -> tuple[T, T]:
    if isinstance(x, Iterable):
        return x  # type: ignore
    return (x, x)


def truncated_mean(
    scores: np.ndarray,
    limits: tuple[float, float] | None = None,
    top_limit=2.0,
    bottom_limit=float("inf"),
    base="tukey",
    return_limits=False,
) -> float | tuple[float, tuple[float, float] | None]:
    """
    Calculates the truncated mean

    Args:
        scores: A list of scores for which to calculate the truncated mean
        limits: The limits to use for trimming the scores. If not given, they will
                be calculated based on `top_limit`, `bottom_limit` and `base.`
                You can pass the special value "mean" as a string to disable limits and
                calcualte the mean. You can pass the special value "median" to calculate
                the median (which is the same as passing top_limit==bottom_limit=0.5
                and base == "scores").
        top_limit: top limit on scores to use for truncated mean calculation. See `base`
        bottom_limit: bottom limit on scores to use for truncated mean calculation. See `base`
        base: The base for calculating the limits used to apply the `top_limit` and `bottom_limit`.
              Possible values are:
              - zscore: the number of sigmas to remove above/below. A good default choice is 3. Pass inf to disable a side.
              - tukey: the fraction of IQR to remove above/below. A good default choice is 1.5 or 3 (we use 2). Pass inf to disable a side.
              - iqr : same as tukey
              - iqr_fraction: the fraction is interpreted as the fraction of scores above/below the 1st/3rd qauntile
              - scores: the fraction is interpreted as fraction of highest and lowest scores
              - fraction: the fraction is interpreted as literal fraction of the values (i.e. given 10 values and 0.1, removes 1 value)
              - mean: simply returns the mean (limits ignored)
              - median: simply returns the median (limits ignored)
        return_limits: If true, the method will also return the limiting scores used in its mean calculation.
    """

    scores = np.asarray(scores)
    scores = scores[~np.isnan(scores)]

    if isinstance(limits, str) and limits.lower() == "mean":
        return tmean(scores, None) if not return_limits else (tmean(scores, None), None)
    if isinstance(limits, str) and limits.lower() == "median":
        return np.median(scores) if not return_limits else (np.median(scores), None)  # type: ignore (seems ok)
    if limits is not None:
        return np.mean(scores) if not return_limits else (np.mean(scores), None)

    if base == "zscore":
        m, s = np.nanmean(scores), np.nanstd(scores)
        limits_ = (m - s * bottom_limit, m + s * top_limit)
    elif base in ("tukey", "iqr"):
        q1, q3 = np.quantile(scores, 0.25), np.quantile(scores, 0.75)
        iqr = q3 - q1
        limits_ = (
            q1 - (bottom_limit * iqr if not np.isinf(bottom_limit) else bottom_limit),
            q3 + (top_limit * iqr if not np.isinf(top_limit) else top_limit),
        )
    elif base == "iqr_fraction":
        bottom_limit = min(1, max(0, bottom_limit))
        top_limit = min(1, max(0, top_limit))
        limits_: tuple[float, float] = (np.quantile(scores, 0.25), np.quantile(scores, 0.75))  # type: ignore (seems ok)
        high = np.sort(scores[scores > limits_[1]])
        low = np.sort(scores[scores < limits_[0]])
        limits_ = (  # type: ignore
            low[int((len(low) - 1) * bottom_limit)] if len(low) > 0 else None,
            high[int((len(high) - 1) * (1 - top_limit))] if len(high) > 0 else None,
        )
    elif base == "fraction":
        bottom_limit = min(1, max(0, bottom_limit))
        top_limit = min(1, max(0, top_limit))
        scores = np.sort(scores)
        top_indx = int((len(scores) - 1) * (1 - top_limit))
        bottom_indx = int((len(scores) - 1) * bottom_limit)
        if top_indx < bottom_indx:
            return float("nan") if not return_limits else (float("nan"), limits)
        m = np.mean(scores[bottom_indx : top_indx + 1])
        return m if not return_limits else (m, (scores[bottom_indx], scores[top_indx]))
    elif base == "scores":
        bottom_limit = min(1, max(0, bottom_limit))
        top_limit = min(1, max(0, top_limit))
        limits_ = (  # type: ignore
            np.quantile(scores, bottom_limit),
            np.quantile(scores, 1 - top_limit),
        )
        if limits_[0] > limits_[1]:
            return float("nan") if not return_limits else (float("nan"), limits_)
    elif base == "mean":
        return np.mean(scores) if not return_limits else (np.mean(scores), None)
    elif base == "median":
        return np.median(scores) if not return_limits else (np.median(scores), None)  # type: ignore
    else:
        raise ValueError(f"Unknown base for truncated_mean ({base})")
    assert limits_[0] is not None and limits_[1] is not None
    if len(scores) == 0 or limits_[1] < limits_[0]:
        return float("nan") if not return_limits else (float("nan"), limits_)
    try:
        # this is an inclusive trimmed mean
        # tm = tmean(scores, limits)
        scores = scores[scores >= limits_[0]]
        scores = scores[scores <= limits_[1]]
        if len(scores) == 0:
            return float("nan") if not return_limits else (float("nan"), limits_)
        tm = np.mean(scores)
        return tm if not return_limits else (tm, limits_)
    except ValueError:
        return float("nan") if not return_limits else (float("nan"), limits_)


def sample(n, k, grid=False, compact=True, endpoints=True):
    """
    Samples `k` items from `n` in the range (0, `n`-1) optionally explring them

    Args:
        n: The number of items to sample (assumed to range from index 0 to n-1)
        k: The number of samples to take
        grid: Sample on a grid (equally distanced as much as possible)
        compact: If True, the samples will be choosen near each other (see endpoints though)
        endpoints: If given, the first and last index are guaranteed to be in the samples
    """
    if k is None:
        return range(n)
    if n == 0:
        return []
    if k == 1:
        return (0,) if n else []
    if n < 2:
        return (0,) if k else []
    if k > n:
        return range(n)
    pre, post = [], []
    if endpoints:
        pre, post = [0], [n - 1]
        n, k = n - 2, k - 2
    if grid:
        step = int(n // k)
        if step < 1:
            step = 1
        l = 1 + step * ((n - 1) // step)
        l = min(l, n - 1)
        rng = range(step, l, step)
    elif compact:
        before = (n - k) // 2
        rng = range(before, min(n, before + k))
    else:
        rng = random.sample(range(n), k)
    if not endpoints:
        return rng
    return pre + list(rng) + post



from __future__ import annotations

import time

import pytest

from negmas.outcomes.base_issue import make_issue
from negmas.preferences import LinearAdditiveUtilityFunction as LUFun
from negmas.preferences.crisp.nonlinear import HyperRectangleUtilityFunction
from negmas.preferences.value_fun import AffineFun, IdentityFun, LinearFun
from negmas.sao.mechanism import SAOMechanism
from negmas.sao.negotiators import AspirationNegotiator, NaiveTitForTatNegotiator
from negmas.sao.negotiators.limited import LimitedOutcomesNegotiator


def test_a_session():
    time.perf_counter()
    n = 50
    p = SAOMechanism(outcomes=n, n_steps=50)
    for _ in range(4):
        p.add(
            LimitedOutcomesNegotiator(p_ending=0.01, name=f"agent {_}"),
            preferences=HyperRectangleUtilityFunction(
                [None], [lambda x: x[0]], outcomes=((_,) for _ in range(n))
            ),
        )
    p.run()
    # print(f'{len(p.negotiators)} negotiators')
    assert len(p.history) > 0
    # print(f'Took {time.perf_counter()-start}')


@pytest.mark.skip(
    "Will check this later. Someone breaks the negotiation and there is a difference based on utility normalizerion!!"
)
def test_buy_sell_session():
    # create negotiation agenda (issues)
    issues = [
        make_issue(name="price", values=10),
        make_issue(name="quantity", values=(1, 11)),
        make_issue(name="delivery_time", values=10),
    ]

    # create the mechanism
    session = SAOMechanism(issues=issues, n_steps=20)

    # define buyer and seller utilities
    seller_utility = LUFun(
        values=[IdentityFun(), LinearFun(0.2), AffineFun(-1, bias=9.0)],
        outcome_space=session.outcome_space,
    ).scale_max(1.0)

    buyer_utility = LUFun(
        values={
            "price": AffineFun(-1, bias=9.0),
            "quantity": LinearFun(0.2),
            "delivery_time": IdentityFun(),
        },
        outcome_space=session.outcome_space,
    ).scale_max(1.0)

    # create and add buyer and seller negotiators
    session.add(NaiveTitForTatNegotiator(name="buyer"), preferences=buyer_utility)
    session.add(AspirationNegotiator(name="seller"), preferences=seller_utility)

    # run the negotiation and show the results
    state = session.run()
    import matplotlib.pyplot as plt

    assert state.agreement is not None, f"{neg.plot()}{plt.show()}{neg.extended_trace}"


if __name__ == "__main__":
    pytest.main(args=[__file__])



from __future__ import annotations

import hypothesis.strategies as st
from hypothesis import example, given, settings
from pytest import mark

import negmas
from negmas import all_negotiator_types
from negmas.outcomes import Issue, make_issue
from negmas.preferences import LinearAdditiveUtilityFunction
from negmas.sao import EndImmediately, NoneOfferingStrategy, RejectAlways, SAOMechanism
from negmas.sao.negotiators.modular.boa import make_boa

NEGTYPES = all_negotiator_types()


@given(
    opp=st.sampled_from(NEGTYPES),
    start=st.booleans(),
    rejector=st.sampled_from([EndImmediately, RejectAlways]),
    avoid_ultimatum=st.booleans(),
)
@example(
    opp=negmas.sao.negotiators.timebased.AdditiveFirstFollowingTBNegotiator,
    start=True,
    rejector=negmas.sao.components.acceptance.EndImmediately,
    avoid_ultimatum=True,
)
@settings(deadline=500000)
def test_do_nothing_never_gets_agreements(opp, start, rejector, avoid_ultimatum):
    agent = make_boa(acceptance=rejector(), offering=NoneOfferingStrategy())
    issues: list[Issue] = [
        make_issue(10, "price"),
        make_issue(10, "quantity"),
        make_issue(["red", "green", "blue"], "color"),
    ]
    ufuns = [
        LinearAdditiveUtilityFunction.random(issues=issues),
        LinearAdditiveUtilityFunction.random(issues=issues),
    ]
    session = SAOMechanism(n_steps=1000, issues=issues, avoid_ultimatum=avoid_ultimatum)
    negs = [opp(), agent] if not start else [agent, opp()]
    for n, u in zip(negs, ufuns):
        session.add(n, preferences=u)

    assert session.run().agreement is None


@mark.parametrize(
    ["factory", "name", "short_name"], [(make_boa, "BOANegotiator", "BOA")]
)
def test_has_correct_type_name(factory, name, short_name):
    x = factory()
    assert x.type_name == name
    assert x.short_type_name == short_name



from __future__ import annotations

import math
import numbers
import random
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable

import numpy as np

from negmas import warnings
from negmas.helpers.numeric import sample
from negmas.helpers.strings import unique_name
from negmas.helpers.types import get_full_type_name
from negmas.protocols import HasMinMax
from negmas.serialization import PYTHON_CLASS_IDENTIFIER, deserialize, serialize

if TYPE_CHECKING:
    from .common import Outcome

__all__ = ["make_issue", "Issue", "DiscreteIssue"]


def make_issue(values, *args, **kwargs):
    """
    A factory for creating issues based on `values` type as well as the base class of all issues

    Args:
            values: Possible values for the issue
            name: Name of the issue. If not given, a random name will be generated
            id: The ID of the issue. Should be unique in the whole system.

    Remarks:

        - Issues can be initialized by either an iterable of strings, an integer or a tuple of two values with
          the following meanings:

          - ``list of anything`` : This is an issue that can any value within the given set of values
            (strings, ints, floats, etc). Depending on the types in the list, a different issue type will be created:

                - integers -> CardinalIssue
                - a type that supports subtraction -> OrdinalIssue (with defined order and defined difference between values)
                - otherwise -> CategoricalIssue (without defined order or difference between values)

          - ``int`` : This is a ContiguousIssue that takes any value from 0 to the given value -1 (int)
          - Tuple[ ``int`` , ``int`` ] : This is a ContiguousIssue that can take any integer value in the given limits (min, max)
          - Tuple[ ``float`` , ``float`` ] : This is a ContinuousIssue that can take any real value in the given limits (min, max)
          - Tuple[ ``int`` , ``inf`` ] : This is a CountableInfiniteIssue that can take any integer value in the given limits
          - Tuple[ ``-inf`` , ``int`` ] : This is a CountableInfiniteIssue that can take any integer value in the given limits
          - Tuple[ ``float`` , ``inf`` ] : This is a ContinuousInfiniteIssue that can take any real value in the given limits
          - Tuple[ ``-inf`` , ``float`` ] : This is a ContinuousInfiniteIssue that can take any real value in the given limits
          - ``Callable`` : The callable should take no parameters and should act as a generator of issue values. This
             type of issue is always assumed to be neither countable nor continuous and are called uncountable. For
             example, you can use this type to make an issue that generates all integers from 0 to infinity. Most operations are not
             supported on this issue type.
        - If a list is given, min, max must be callable on it.
    """
    from negmas.outcomes.callable_issue import CallableIssue
    from negmas.outcomes.cardinal_issue import DiscreteCardinalIssue
    from negmas.outcomes.categorical_issue import CategoricalIssue
    from negmas.outcomes.contiguous_issue import ContiguousIssue
    from negmas.outcomes.continuous_issue import ContinuousIssue
    from negmas.outcomes.infinite import ContinuousInfiniteIssue, CountableInfiniteIssue
    from negmas.outcomes.ordinal_issue import DiscreteOrdinalIssue

    if isinstance(values, numbers.Integral):
        return ContiguousIssue(int(values), *args, **kwargs)
    if isinstance(values, tuple):
        if len(values) != 2:
            raise ValueError(
                f"Passing {values} is illegal. Issues with ranges need 2-values tuples"
            )
        if isinstance(values[0], numbers.Integral) and isinstance(
            values[1], numbers.Integral
        ):
            return ContiguousIssue(values, *args, **kwargs)  # type: ignore (we know that the types are OK here)
        if (
            isinstance(values[0], numbers.Integral)
            and values[1] == float("inf")
            or isinstance(values[1], numbers.Integral)
            and values[0] == float("-inf")
        ):
            return CountableInfiniteIssue(values, *args, **kwargs)  # type: ignore (we know that the types are OK here)
        if (
            isinstance(values[0], numbers.Real)
            and values[1] == float("inf")
            or isinstance(values[1], numbers.Real)
            and values[0] == float("-inf")
        ):
            return ContinuousInfiniteIssue(values, *args, **kwargs)
        if isinstance(values[0], numbers.Real) and isinstance(values[1], numbers.Real):
            return ContinuousIssue(values, *args, **kwargs)
        raise ValueError(
            f"Passing {values} with mixed types. Both values must be either integers or reals"
        )
    if isinstance(values, Callable):
        return CallableIssue(values, *args, **kwargs)  # type: ignore
    if isinstance(values, Iterable) and all(
        isinstance(_, numbers.Integral) for _ in values
    ):
        return DiscreteCardinalIssue(values, *args, **kwargs)
    if isinstance(values, Iterable):
        values = list(values)
        try:
            for a, b in zip(values[1:], values[:-1]):
                a - b
        except:
            return CategoricalIssue(values, *args, **kwargs)
        return DiscreteOrdinalIssue(values, *args, **kwargs)
    return CategoricalIssue(values, *args, **kwargs)


class Issue(HasMinMax, Iterable, ABC):
    """
    Base class of all issues in NegMAS
    """

    def __init__(
        self,
        values,
        name: str | None = None,
    ) -> None:
        self.name = name if name else unique_name("issue", add_time=False, sep="")
        self._value_type = object
        self._values = values
        self._n_values = float("inf")
        self.min_value, self.max_value = None, None

    def __copy__(self):
        return make_issue(name=self.name, values=self._values)

    def __deepcopy__(self, memodict={}):
        return make_issue(name=self.name, values=self._values)

    @property
    def value_type(self):
        """
        Returns the type of values in this issue
        """
        return self._value_type

    @property
    def values(self):
        """
        Returns the raw values representation of the issue. Only use if you know what you are doing. To get all the values that can be assigned to this issue use `all` or `generate_values`
        """
        return self._values

    def has_limits(self) -> bool:
        """
        Checks whether the minimum and maximum values of the issue are known
        """
        return self.min_value is not None and self.max_value is not None

    def is_numeric(self) -> bool:
        """
        Checks that each value of this issue is a number
        """
        return issubclass(self._value_type, numbers.Number)

    def has_finite_limits(self) -> bool:
        """
        Checks whether the minimum and maximum values of the issue are known and are finite
        """
        return (
            self.has_limits()
            and self.is_numeric()
            and math.isfinite(self.min_value)
            and math.isfinite(self.max_value)
        )

    def is_integer(self) -> bool:
        """
        Checks that each value of this issue is an integer
        """
        return issubclass(self._value_type, numbers.Integral)

    def is_float(self) -> bool:
        """
        Checks that each value of this issue is a real number
        """
        return issubclass(self._value_type, numbers.Real) and not issubclass(
            self._value_type, numbers.Integral
        )

    @abstractmethod
    def is_continuous(self) -> bool:
        """
        The issue has a continuous set of values. Note that this is different from having values that are real (which is tested using `is_float` )
        """
        ...

    def is_discrete(self) -> bool:
        """
        Checks whether the issue has a discrete set of values. This is different from `is_integer` which checks that the values themselves are integers and `is_discrete_valued` which checks that they are discrete.
        """
        return not self.is_continuous()

    def is_finite(self) -> bool:
        """
        Checks whether the issue has a discrete set of values
        """
        return self.is_discrete()

    def is_discrete_valued(self) -> bool:
        """
        Checks that each value of this issue is not a real number
        """
        return not self.is_float()

    @property
    def cardinality(self) -> int | float:
        """The number of possible outcomes for the issue. Returns infinity for continuous and uncountable spaces"""
        return self._n_values

    @abstractmethod
    def rand(self) -> int | float | str:
        """Picks a random valid value."""

    def rand_valid(self):
        """
        Generates a random  valid value for this issue
        """
        return self.rand()

    @classmethod
    def from_dict(cls, d):
        """
        Constructs an issue from a dict generated using `to_dict()`
        """
        if isinstance(d, cls):
            return d
        d.pop(PYTHON_CLASS_IDENTIFIER, None)
        d["values"] = deserialize(d["values"])
        return cls(
            values=d.get("values", None),
            name=d.get("name", None),
        )

    def to_dict(self):
        """
        Converts the issue to a dictionary from which it can be constructed again using `Issue.from_dict()`
        """
        d = {PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self))}
        return dict(
            **d,
            values=serialize(self.values),
            name=self.name,
            n_values=self._n_values,
        )

    @abstractmethod
    def is_valid(self, v):
        """Checks whether the given value is valid for this issue"""
        ...

    def contains(self, issue: Issue) -> bool:
        """
        Checks weather this issue contains the input issue (i.e. every value in the input issue is in this issue)
        """
        if isinstance(issue, DiscreteIssue):
            return all(self.is_valid(_) for _ in issue.all)
        return False

    @property
    def type(self) -> str:
        """
        Returns a nice name for the issue type
        """
        return self.__class__.__name__.lower().replace("issue", "")

    @abstractmethod
    def ordered_value_generator(
        self, n: int | float | None = None, grid=True, compact=False, endpoints=True
    ) -> Generator[int, None, None]:
        """
        A generator that generates at most `n` values (in a stable order)


        Args:
            n: The number of samples. If inf or None, all values will be generated but when the issue is infinite, it will just fail
            grid: Sample on a grid (equally distanced as much as possible)
            compact: If True, the samples will be choosen near each other (see endpoints though)
            endpoints: If given, the first and last index are guaranteed to be in the samples

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If the order is not defined for this issue, this generator will still generate values in the same order every time it is called.
            - If you need a list then use something like:

            >>> from negmas.outcomes import make_issue
            >>> list(make_issue(5).value_generator())
            [0, 1, 2, 3, 4]
            >>> list(int(10 * _) for _ in make_issue((0.0, 1.0)).value_generator(11))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        """

    @abstractmethod
    def value_generator(
        self, n: int | float | None = 10, grid=True, compact=True, endpoints=True
    ) -> Generator[Any, None, None]:
        """
        A generator that generates at most `n` values (in any order)


        Args:
            grid: Sample on a grid (equally distanced as much as possible)
            compact: If True, the samples will be choosen near each other (see endpoints though)
            endpoints: If given, the first and last index are guaranteed to be in the samples

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import make_issue
            >>> list(make_issue(5).value_generator())
            [0, 1, 2, 3, 4]
            >>> list(int(10 * _) for _ in make_issue((0.0, 1.0)).value_generator(11))
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

        """

    def to_discrete(
        self, n: int | float | None = 10, grid=True, compact=True, endpoints=True
    ) -> DiscreteIssue:
        """
        Converts the issue to a discrete issue by samling from it.

        If the issue is already discret it will just return itself. This method cannot be used to reduce the cardinality of a discrete issue.

        Args:
            n (int | float | None): Number of values in the resulting discrete issue. This will be ignored if the issue is already discrete. The only allowed float value is `float("inf")`. If any other float is passed, it will be silently cast to an int
            grid (bool): Sample on a grid
            compact (bool): Sample around the center
            endpoints (bool): Always incllude minimum and maximum  values

        """
        from negmas.outcomes.categorical_issue import CategoricalIssue

        if isinstance(self, DiscreteIssue):
            return self

        return CategoricalIssue(
            list(self.value_generator(n, grid, compact, endpoints)),
            name=self.name,
        )

    @abstractmethod
    def _to_xml_str(self, indx: int) -> str:
        ...

    @abstractmethod
    def value_at(self, index: int):
        """
        Returns the value at the given indes of the issue. The same index  will have the same values always indepdendent of whether the values of the issue have defined ordering.
        """
        ...

    @abstractmethod
    def rand_outcomes(
        self, n: int, with_replacement=False, fail_if_not_enough=False
    ) -> list:
        """
        Picks n random valid value (at most).

        Args:
            n: The number of outcome values to sample
            with_replacement: If true, sampling is done with replacement (i.e.repetition is allowed)
            fail_if_not_enough: If true, raises an exception if it is not possible to sample exactly
                                `n` values. If false, will sample as many values as possible up to `n`
        Returns:
            A list of sampled values
        """

    @abstractmethod
    def rand_invalid(self):
        """Pick a random *invalid* value"""

    @property
    def all(self) -> Generator[Any, None, None]:
        """
        A generator that generates all possible values.
        """
        if self.is_discrete():
            return self.value_generator()
        raise ValueError(
            f"The issue ({self}) is not discrete and `all` cannot be called on it"
        )

    def __getitem__(self, indx):
        return self.value_at(indx)

    def __iter__(self):
        return self.value_generator().__iter__()

    def __contains__(self, item):
        try:
            if isinstance(item, Issue):
                return self.contains(item)
            return self.is_valid(item)
        except Exception as e:
            warnings.warn(
                f"Testing whether {item} is contained  in {self} threw an exception: {e}. continuing as if it is not",
                warnings.NegmasCaughtExceptionWarning,
            )
        return False

    def __len__(self):
        return self.cardinality

    def __eq__(self, other):
        return self._values == other.values and self.name == other.name

    def __hash__(self):
        return hash(str(self))

    def __repr__(self):
        return f"Issue({self._values}, {self.name})"

    def __str__(self):
        return f"{self.name}: {self._values}"


class DiscreteIssue(Issue):
    """
    An `Issue` with a discrete set of values.
    """

    @property
    def cardinality(self) -> int:
        """The number of possible outcomes for the issue. Guaranteed to  be fininte"""
        return self._n_values  # type: ignore

    def is_continuous(self) -> bool:
        return False

    @property
    @abstractmethod
    def all(self) -> Generator[Any, None, None]:
        """
        A generator that generates all possible values.

        Remarks:
            - This function returns a generator for the case when the number of values is very large.
            - If you need a list then use something like:

            >>> from negmas.outcomes import make_issue
            >>> list(make_issue(5).all)
            [0, 1, 2, 3, 4]

        """

    def ordered_value_generator(
        self, n: int | float | None = 10, grid=True, compact=True, endpoints=True
    ) -> Generator[Any, None, None]:
        m = self.cardinality
        n = m if n is None or not math.isfinite(n) else int(n)

        for i in range(n):
            yield self._values[i % m]

    def value_generator(
        self, n: int | float | None = 10, grid=True, compact=True, endpoints=True
    ) -> Generator[Any, None, None]:
        m = self.cardinality
        n = m if n is None or not math.isfinite(n) else int(n)

        yield from (
            self._values[_]
            for _ in sample(m, n, grid=grid, compact=compact, endpoints=endpoints)
        )

    def value_at(self, index: int):
        if index < 0 or index > self.cardinality - 1:
            raise IndexError(index)
        return self._values[index]

    def rand(self):
        """Picks a random valid value."""
        return random.choice(self._values)  # type: ignore

    def rand_outcomes(
        self, n: int, with_replacement=False, fail_if_not_enough=False
    ) -> Iterable[Outcome]:
        """Picks a random valid value."""

        if n > len(self._values) and not with_replacement:
            if fail_if_not_enough:
                raise ValueError(
                    f"Cannot sample {n} outcomes out of {self._values} without replacement"
                )
            else:
                return self._values

        return np.random.choice(
            np.asarray(self._values, dtype=self._value_type),
            size=n,
            replace=with_replacement,
        ).tolist()

    def is_valid(self, v):
        return v in self._values

    def __getitem__(self, indx):
        return self.values(indx)



"""
Defines basic concepts related to outcomes

Outcomes in this package are always assumed to be multi-issue outcomes where  single-issue outcomes can be implemented
as the special case with a single issue.

- Both Continuous and discrete issues are supported. All issue will have names. If none is given, a random name will be
  used. It is HIGHLY recommended to always name your issues.
- Outcomes are dictionaries with issue names as keys and issue values as values.

Examples:

  Different ways to create issues:

  >>> issues = [make_issue((0.5, 2.0), 'price'), make_issue(['2018.10.'+ str(_) for _ in range(1, 4)], 'date')
  ...           , make_issue(20, 'count')]
  >>> for _ in issues: print(_)
  price: (0.5, 2.0)
  date: ['2018.10.1', '2018.10.2', '2018.10.3']
  count: (0, 19)

  Outcome example compatible with the given set of issues:

  >>> a = {'price': 1.2, 'date': '2018.10.04', 'count': 4}

"""
from __future__ import annotations


from .common import *
from .protocols import *
from .base_issue import *
from .callable_issue import *
from .categorical_issue import *
from .contiguous_issue import *
from .continuous_issue import *
from .ordinal_issue import *
from .range_issue import *
from .cardinal_issue import *
from .infinite import *
from .issue_ops import *
from .outcome_ops import *
from .outcome_space import *

__all__ = (
    common.__all__
    + protocols.__all__
    + base_issue.__all__
    + callable_issue.__all__
    + categorical_issue.__all__
    + ordinal_issue.__all__
    + range_issue.__all__
    + cardinal_issue.__all__
    + contiguous_issue.__all__
    + continuous_issue.__all__
    + infinite.__all__
    + issue_ops.__all__
    + outcome_ops.__all__
    + outcome_space.__all__
)
